#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.
from typing import Dict, List
from .species import Species, ComplexSpecies
from .polymer import OrderedPolymer, OrderedMonomer, NamedPolymer
from .dna_construct import Construct, DNA_construct
from .dna_part_misc import IntegraseSite
from .component_enumerator import GlobalComponentEnumerator
import itertools as it
from .utils import combine_dictionaries
import copy
class Polymer_transformation:
    def __init__(self,partslist,circular=False,parentsdict = None,material_type="dna"):
        """A Polymer transformation is like a generic transformation of a polymer sequence.
        You specify a parts list that would make up the output polymer. This list can contain:
        parts from ordered polymers
        parts that aren't in any polymers (have parent = None)
        parts from ordered polymers are considered as "placeholders"
        parts with parent = None are inserted into the new polymer.

        Also you can specify if the output should be circular or not.

        example:

        valid partslist:
        partslist = [Monomer(forward,3,"input1"),Monomer(reverse,1,"input2"),Promoter(forward,None,None)]
        
        then, self.create_polymer([polymer1,polymer2])
        takes element 3 from polymer1 and puts it forward, element 1 from polymer 2, and a Promoter object and creates
        a new polymer by feeding these three monomers into a polymer constructor.

        new_polymer.parts_list = [polymer1[3].setdir("forward",polymer2[1].setdir("reverse"),[Promoter,"forward"]]
        """
        if(parentsdict is None):
            parentsdict = {} #the input to this function is a list of monomers that belong to various parents.
                             #each different parent that is represented in these monomers is converted into "input#" arbitrarily
                             #to keep it consistant, a link of that parent to the appropriate "input#" is kept in this dictionary.
                             #you can also pass a pre-populated dictionary into this function in order to control which parent gets
                             #which "input#". This is essential if we want to properly "reverse" a transformation (input1 becomes input2 and input2 becomes input1)
        inputcount = 1
        actual_partslist = []
        partdir = -1 #-1 = not set. Valid values are "forward" "reverse" and None
        #go through the parts
        inputname = None
        new_parentsdict = {}
        for part in partslist:
            if(type(part)==list or type(part) ==tuple):
                #if the part is a list that means it looks like [OrderedMonomer,"direction"]
                part_ref = part[0]
                partdir = part[1]
            else:
                part_ref = part
                partdir = part_ref.direction
            #if the parent is populated, it means this part should be a placeholder
            if(part_ref.parent is not None):
                #if we haven't tracked it already
                if((part_ref.parent not in new_parentsdict) and (part_ref.parent not in parentsdict)):
                    #create a 'blank' polymer based on the input, and give it a generic name
                    inputname = "input"+str(inputcount)
                    while(inputname in list(new_parentsdict.values())):
                        inputcount+=1
                        inputname = "input"+str(inputcount)
                    dummyPolymer = self.dummify(part_ref.parent,inputname)
                    #set the value in the dictionary
                    new_parentsdict[part_ref.parent]=dummyPolymer.name
                    #increment for the next time this happens
                    inputcount+=1
                    
                elif((part_ref.parent not in new_parentsdict) and (part_ref.parent in parentsdict)):
                    inputname = parentsdict[part_ref.parent]
                    dummyPolymer = self.dummify(part_ref.parent,inputname)
                    new_parentsdict[part_ref.parent]=dummyPolymer.name
                elif(inputname is None or inputname != new_parentsdict[part_ref.parent]):
                    inputname = new_parentsdict[part_ref.parent]
                    dummyPolymer = self.dummify(part_ref.parent,inputname)
                #this variable is the actual parts list that will be stored. It has mostly dummy parts
                if(part_ref.parent[part_ref.position]!=part_ref):
                    #this will be true only for situations where a component needs to be transformed into
                    #another component but still needs to keep a reference to where it was in the parent
                    copied_part = part_ref.get_orphan()
                    copied_part.parent = dummyPolymer
                    actual_partslist += [[copied_part,partdir]]
                else:
                    actual_partslist += [[dummyPolymer[part_ref.position],partdir]]
            else:
                #if the part has no parent, copy it and put it in
                actual_partslist += [[copy.deepcopy(part_ref),partdir]]
        self.number_of_inputs = len(new_parentsdict)
        self.parentsdict = new_parentsdict
        self.partslist = actual_partslist
        self.circular = circular
        self.material_type = material_type
    def renumber_output(self,output_renumbering_function):
        """change the ordering of the output list, using the output_renumbering_function which takes
        in an int and returns an int which is the new index of the part"""
        new_partslist = []
        for i in range(len(self.partslist)):
            new_i,direc = output_renumbering_function(i)
            part = self.partslist[new_i]
            if(direc == "r"):
                new_partslist+= [[part[0],["forward","reverse"][part[1]=="forward"]]]
            else:
                new_partslist+= [part]
        self.partslist = new_partslist

    def get_renumbered(self,output_renumbering_function):
        """return a copy of this transformation, but the output indexes are renumbered"""
        rxn_copied = copy.copy(self)
        rxn_copied.renumber_output(output_renumbering_function)
        return rxn_copied
    def reversed(self):
        """return a circularly permuted version of self. That means the inputs are shuffled around
        For example, we had input1, input2, input3. Now we will have input1=input2, input2=input3, input3=input1."""
        new_parentsdict = {}

        new_name_dict = {}
        for inputnum in range(self.number_of_inputs):
            if(inputnum == self.number_of_inputs-1):
                new_name_dict["input"+str(inputnum+1)]="input1"
            else:
                new_name_dict["input"+str(inputnum+1)]="input"+str(inputnum+2)
            
        if(len(self.parentsdict)==1):
            #in this case we have only "input1"
            return self
        else:
            for part in self.partslist:
                if(part[0].parent is not None and part[0].parent not in new_parentsdict):
                    #this next line outputs input2 when given input1
                    
                    newname = new_name_dict[part[0].parent.name]
                    new_parentsdict[part[0].parent]=newname
            return Polymer_transformation(self.partslist,self.circular,parentsdict=new_parentsdict)
    def create_polymer(self,polymer_list,**keywords):
        """this function creates a new polymer from the template saved inside this class.
        A polymer_list is a list of polymers from which the resulting polymer is made. Some of
        the parts which compose the output polymer don't have a parent, and therefore are new parts.
        In these cases anything bound to the previous location of these parts will be bound to the new ones as well."""
        polymer_dict = {"input"+str(a+1):b for a,b in enumerate(polymer_list)}
        assert(len(polymer_list)>=self.number_of_inputs)
        outlst = []
        for part_list in self.partslist:
            part = part_list[0]
            partdir = part_list[1]
            outpart = None
            
            if((part.parent is not None) and not hasattr(part,"name")):
                outpart = polymer_dict[part.parent.name][part.position] #grab the part from the proper input
            elif((part.parent is not None) and hasattr(part,"name")):
                #in this case we are transforming a part into a different part, taking the complexes from
                #the previous position
                
                if(isinstance(polymer_dict["input1"],Construct)):
                    outpart = part.get_orphan()
                else:
                    template_part = polymer_dict[part.parent.name][part.position]
                    if(isinstance(template_part,ComplexSpecies)):
                        #in this case we have to replace the thing inside the complex with the new part, but leave
                        #everything else the same.
                        old_species = template_part.get_species(recursive=True)
                        core_parts = []
                        for spec in old_species:
                            if(spec.material_type=="part" and not spec == template_part):#if you have a material type of part then you are
                                                           #part of a polymer and thus will be replaced
                                core_parts += [spec]
                        #now we should have found only one core part. If there are multiple core parts then
                        #we might have to get crazy.
                        new_part = part.dna_species
                        if(len(core_parts)==1):
                            new_part = template_part.replace_species(core_parts[0],part.dna_species)
                        else:
                            raise KeyError(f"{template_part} contained more than one species with material_type=\"part\", they were {core_parts}")
                        outpart = new_part
                    else:
                        outpart = part.dna_species
            else:
                #parts that don't come from input1 or input2.
                #they need to be either DNA_part or species objects, depending on
                #what kind of object we are making in the end.
                if(isinstance(polymer_dict["input1"],Construct)):
                    outpart = part
                else:
                    outpart = part.dna_species
            #assuming the stored parts have a valid direction
            if(hasattr(outpart,"linked_sites")):
                outpart = copy.copy(outpart)
                outpart.linked_sites = {} #make sure that any integrase sites we copy this way have no
                                            #linked sites, as those would not be links created by the integrate() function
            outlst += [[outpart,partdir]]
        if(hasattr(outlst[0][0],"material_type") and any(["complex" in a[0].material_type for a in outlst])):
            outpolymer = polymer_dict["input1"].__class__(outlst,circular = self.circular,material_type = "ordered_polymer")
        else:
            outpolymer = polymer_dict["input1"].__class__(outlst,circular = self.circular,material_type=self.material_type)
        return outpolymer
    @classmethod
    def dummify(cls,in_polymer,name):
        """creates a simplified polymer that has the same number of monomers, direction of monomers,
        and name as the input polymer, but is otherwise disconnected."""
        #this is used specifically with polymerTransformation. Dummified version of polymers are stored in
        #polymerTransformation as generic "slots" for monomers to be properly placed into
        out_list = []    
        for element in in_polymer:
            out_list += [OrderedMonomer(direction=element.direction)]
        circular = False
        if(hasattr(in_polymer,"circular")):
            circular = in_polymer.circular
        return NamedPolymer(out_list,name,circular=circular)
    def __repr__(self):
        part_texts = []
        for plist in self.partslist:
            part = plist[0]
            part_dir = plist[1]
            if(part.parent is not None):
                if(part.parent[part.position]!=part):
                    part_texts += [[part.name,part.position,part_dir]]
                else:
                    part_texts += [[part.parent.name,part.position,part_dir]]
            else:
                part_texts += [[part.name,part_dir]]
        out_txt = "Polymer transformation = "
        for part_text in part_texts:
            out_txt += "("+str(part_text[0])
            if(len(part_text)==3):
                out_txt +="-"+str(part_text[1])
            if(part_text[-1]!="forward"):
                out_txt += "-r"
            out_txt += ")"
        if(self.circular):
            out_txt+="(circular)"
        return out_txt

class IntegraseRule:
    def __init__(self,name=None,reactions=None,allow_deletion=True,allow_integration=True,allow_inversion=True):
        """The integrase mechanism is a mechanism at the level of DNA. It creates DNA species which
        the integrase manipulations would lead to. This mechanism does not create any reaction rates.
        We need to figure out how integrase binding will work before being able to create
        reactions and their corresponding rates"""
        if(reactions is None):
            reactions = {("attB","attP"):"attL",("attP","attB"):"attR"}
        if(name is None):
            self.name = "int1"
        else:
            self.name = name
        self.allow_deletion = allow_deletion
        self.allow_integration = allow_integration
        self.allow_inversion = allow_inversion
        self.integrase_species = Species(self.name,material_type="protein")
        self.reactions = reactions
        self.attsites = []
        for reaction in reactions:
            self.attsites+=list(reaction)+[reactions[reaction]]
        self.attsites = list(set(self.attsites))
        self.integrations_to_do = [] #these are the reactions that will be performed at compile time
    def binds_to(self):
        return self.attsites
    def reaction_allowed(self,site1,site2):
        assert(isinstance(site1,IntegraseSite))
        assert(isinstance(site2,IntegraseSite))
        assert site1.integrase==site2.integrase
        assert site1.integrase==self.integrase_species
        if(tuple([site1.site_type,site2.site_type]) in self.reactions):
            return True
        return False
    def reactive_sites(self):
        """returns a list of attachment sites (strings) that participate in integrase reactions"""
        attsites = []
        for reaction in self.reactions:
            attsites+=list(reaction)
        attsites = list(set(attsites))
        return attsites
    def generate_products(self,site1,site2,site2_parent=None):
        """generates DNA_part objects corresponding to the products of recombination"""
        #the sites should have the same integrase and dinucleotide, otherwise it won't work
        assert (site1.integrase == site2.integrase)
        assert(site1.integrase==self.integrase_species), f"sites have integrase {site1.integrase} but should be {self.integrase_species}"
        assert(site1.dinucleotide == site2.dinucleotide)
        integrase = site1.integrase
        dinucleotide = site1.dinucleotide
        rxn1 = (site1.site_type,site2.site_type)
        rxn2 = (site2.site_type,site1.site_type)
        #this next part checks if these parts can even react
        if(rxn1 in self.reactions):
            prod1 = self.reactions[rxn1]
        else:
            raise KeyError("{} not found to react with {} in {}".format(site1,site2,self.reactions))
        
        if(rxn2 in self.reactions):
            prod2 = self.reactions[rxn2]
        else:
            raise KeyError("{} not found to react with {} in {}".format(site2,site1,self.reactions))
            
        
        part_prod1 = IntegraseSite(prod1,prod1,dinucleotide=dinucleotide,integrase=integrase,
                            direction=site1.direction,integrase_binding=site1.integrase_binding,
                            material_type = site1.material_type)
        
        part_prod2 = IntegraseSite(prod2,prod2,dinucleotide=dinucleotide,integrase=integrase,
                            direction=site2.direction,integrase_binding=site1.integrase_binding,
                            material_type = site2.material_type)
        if(site1.direction=="forward"):
            part_prod1.position = site1.position
            part_prod1.parent = site1.parent
            part_prod2.position = site2.position
            if(site2_parent is not None):
                part_prod2.parent = site2_parent
            else:
                part_prod2.parent = site2.parent
            return (part_prod1,part_prod2)
        else:
            part_prod2.direction = site1.direction
            part_prod2.position = site1.position
            part_prod2.parent = site1.parent
            part_prod1.direction = site2.direction
            part_prod1.position = site2.position
            if(site2_parent is not None):
                part_prod1.parent = site2_parent
            else:
                part_prod1.parent = site2.parent
            
            return(part_prod2,part_prod1)
    
    def integrate(self,site1,site2,also_inter=True,force_inter=False, existing_dna_constructs = None):
        """perform an integration reaction between the chosen sites and make new DNA_constructs
        site1 and site2 are integrase site dna_parts which have parents that are DNA_constructs.
        
        There are four possible reactions:
        1) inversion
           two sites are part of the same dna construct
           the result is another dna construct with the same circularity and the region in between the sites flipped
        2) deletion
           two sites are part of the same dna construct
           the result is two dna constructs: one with the same circularity but the region between the sites deleted, and another
           circular dna construct that contains the deleted portion
        3) integration
           the sites are on two different dna constructs
           the result is a single dna construct
        4) recombination
           the sites are on two different dna constructs
           the results are two different dna constructs with the proper portions swapped
        
        after the correct dna constructs are generated, the reactions which were done to produce them
        are encoded into polymer_transformations and "baked into" the integrase sites themselves. So,
        each integrase site knows which specific integrase reactions it should produce when it comes
        time to update_reactions.
        
        also_inter controls whether intramolecular reactions should also generate intermolecular reactions
        that occur between two copies of the same plasmid.

        force_inter forces a reaction to be intermolecular even if the two sites are on the same plasmid
        """
        intermolecular = True #by default, the reaction is intermolecular
        #if one of the sites is not part of a construct then raise an error!
        integ_funcs = []
        new_dna_constructs = [] #new dna constructs made by this function!
        if(existing_dna_constructs is None):
            existing_dna_constructs = []
        if(not(isinstance(site1.parent,Construct))):
            raise ValueError("{} not part of a construct".format(site1))
        elif(not(isinstance(site2.parent,Construct))):
            raise ValueError("{} not part of a construct".format(site2))
        cutpos1 = site1.position
        cutpos2 = site2.position
        #below are the references to the sites in the products
        dna_inputs = []
        if(site1.parent==site2.parent and not force_inter):
            intermolecular = False
            #these sites are part of the same piece of DNA, so they are going to do an intramolecular reaction
            contains_no_inter = any(['no_inter' in a.attributes for a in site1.parent])
            if(also_inter and not contains_no_inter):
                #we should generate the intermolecular reaction also!
                #in this case we are generating multiple integration reactions at the same time
                #i think the right thing to do is NOT return it?
                new_dna_constructs += self.integrate(site1,site2,force_inter=True,existing_dna_constructs = existing_dna_constructs)
            if(site1.position > site2.position):
                #we reverse which position is where
                cutpos2 = site1.position
                cutpos1 = site2.position
                prod1,prod2 = self.generate_products(site2,site1)
                #integrase sites are converted into different sites according to this function
            else:
                prod1,prod2 = self.generate_products(site1,site2)
                #integrase sites are converted into different sites according to this function
            
            dna = site1.parent
            dna_inputs = [dna]
            circularity = dna.circular
            
            if(site1.direction == site2.direction):
                if(self.allow_deletion):
                    #case 2: deletion
                    #if the sites point in the same direction, then we are doing a deletion reaction
                    #direction doesn't matter so we don't need to flip anything
                    cutdna_list_parts = list(dna[:cutpos1])+[[prod1,dna[cutpos1].direction]]+list(dna[cutpos2+1:]) #delete
                    newdna_list_parts = [[prod2,dna[cutpos2].direction]]+list(dna[1+cutpos1:cutpos2])

                    integ_funcs += [Polymer_transformation(cutdna_list_parts,circular = circularity),\
                                            Polymer_transformation(newdna_list_parts,circular=True)]
            else:
                if(self.allow_inversion):
                    #case 1: inversion
                    #this means we are dealing with an inversion
                    inv_segment = []
                    
                    [[a,] for a in dna[cutpos1+1:cutpos2][::-1]]
                    for a in dna[cutpos1+1:cutpos2][::-1]:
                        inv_segment += [[a,["forward","reverse"][a.direction=="forward"]]]
                    #the inverted segment is reversed
                    
                    invertdna_list = list(dna[:cutpos1])+\
                                    [[prod1,dna[cutpos1].direction]]+\
                                    inv_segment+ \
                                    [[prod2,dna[cutpos2].direction]]+\
                                    list(dna[cutpos2+1:])

                    integ_funcs += [Polymer_transformation(invertdna_list,circular=circularity)]
        else:
            if(self.allow_integration):
                #otherwise these sites are on different pieces of DNA, so they are going to combine
                dna1 = site1.parent
                dna2 = site2.parent
                circ1 = dna1.circular
                circ2 = dna2.circular
                if(dna1 == dna2):
                    #this will happen if we trying to do an intermolecular reaction between two copies of the same thing
                    dna2 = dna1.__class__(dna1.parts_list,dna1.name+"_duplicate",dna1.circular)
                dna_inputs = [dna1,dna2]
                pdict = {a[1]:"input"+str(a[0]+1) for a in enumerate(dna_inputs)}
                #make sure everyone is forwards
                sites = [site1,site2]
                site_halves = []
                for dna_num,site_num in zip(dna_inputs,sites):
                    if(site_num.direction=="reverse"):
                        dnanum_beginning = [[a,["forward","reverse"][a.direction=="forward"]] for a in dna_num[site_num.position+1:][::-1]]
                        dnanum_end = [[a,["forward","reverse"][a.direction=="forward"]] for a in dna_num[:site_num.position][::-1]]
                    else:
                        dnanum_beginning = dna_num[:site_num.position]
                        dnanum_end = dna_num[site_num.position+1:]
                    site_halves += [[list(dnanum_beginning),list(dnanum_end)]]
                dna1_halves = site_halves[0]
                dna2_halves = site_halves[1]

                
                
                if(site1.direction=="reverse" and site2.direction=="reverse"):
                    prod1,prod2 = self.generate_products(site2,site1,site2_parent=dna2)
                else:
                    prod1,prod2 = self.generate_products(site1,site2,site2_parent=dna2)
                #direction of everything should be forward

                if(circ2==True):
                    #case 3: integration
                    #in this case we are combining a circular plasmid with a circular or linear plasmid
                    #either way the result is basically the same, except the result is either linear or circular
                    #result is ONE PIECE OF DNA
                    result = dna1_halves[0]+[[prod1,"forward"]]+dna2_halves[1]+dna2_halves[0]+[[prod2,"forward"]]+dna1_halves[1]
                    integ_funcs += [Polymer_transformation(result,circ1,parentsdict=pdict)]
                elif(circ2 ==False and circ1 == True):
                    #if the sites are backwards just reverse everything
                    new_dna_constructs += self.integrate(site2,site1,force_inter=force_inter,existing_dna_constructs=existing_dna_constructs)
                    #the above already populates the sites, so then we don't need to
                elif(circ1==False and circ1==circ2):
                    #case 4: recombination
                    #here we are recombining two linear dnas, so two linear dnas are produced
                    result1 = dna1_halves[0]+[[prod1,"forward"]]+dna2_halves[1]
                    result2 = dna2_halves[0]+[[prod2,"forward"]]+dna1_halves[1]
                    integ_funcs += [Polymer_transformation(result1,parentsdict=pdict),Polymer_transformation(result2,parentsdict=pdict)]
        if(len(integ_funcs)>0):
            for integ_func in integ_funcs:
                #generate new dna constructs and check if we already made them before
                created_dna = integ_func.create_polymer([site1.parent,site2.parent])
                output = Integrase_Enumerator.find_dna_construct(created_dna,existing_dna_constructs+new_dna_constructs)
                if(output is not None):
                    #if the construct has already been made, then fix the integ_func so it outputs the right thing
                    integ_func.renumber_output(output[1])
                else:
                    #otherwise, add it to the list
                    new_dna_constructs += [created_dna]
            #link the two sites and give them the adjusted integ_funcs
            site1.linked_sites[(site2,intermolecular)] = [integ_funcs,[]]
            site2.linked_sites[(site1,intermolecular)] = [[a.reversed() for a in integ_funcs],[]]
        #the return value of this function is used mostly only for generating constructs
        return new_dna_constructs





class Integrase_Enumerator(GlobalComponentEnumerator):
    def __init__(self,name:str,int_mechanisms = None):
        if(int_mechanisms is None):
            int_mechanisms={"int1":IntegraseRule()}
        self.int_mechanisms = int_mechanisms
        GlobalComponentEnumerator.__init__(self,name=name)
    
    def list_integrase(self,construct):
        """lists all the parts that can be acted on by integrases"""
        int_dict = {}
        for part in construct.parts_list:
            if(isinstance(part,IntegraseSite) and part.integrase is not None):
                part_integrase = part.integrase.name
                if(part_integrase in int_dict):
                    int_dict.update({part_integrase:int_dict[part_integrase]+[part]})
                else:
                    int_dict[part_integrase]=[part]
        return int_dict
    def reset(self,components=None, **keywords):
        """this resets the linked_sites member in any attachment sites"""
        for component in components:
            if(hasattr(component,"parts_list")):
                for part in component:
                    if(hasattr(part,"linked_sites")):
                        part.linked_sites = {}
    @classmethod
    def find_dna_construct(cls,construct:Construct,conlist:List[Construct]):
        """find a construct that matches the input "construct", but can be reverse or circularly permuted
        returns: found_construct, index_function(index) => (new_index,"f" or "r" if it must be reversed)
        or,
        None, if no matching construct is found"""
        matched_construct = None
        for other_construct in conlist:
            if(not isinstance(other_construct,Construct)):
                continue
            if(construct.directionless_hash==other_construct.directionless_hash):
                if(matched_construct is not None):
                    #a construct must not match two constructs, since they are generated and checked in order
                    raise KeyError(f"{construct} matches with {matched_construct} but also {other_construct}")
                matched_construct = other_construct
        if(matched_construct is None):
            return None
        other_construct = matched_construct
        other_indexes = list(range(len(other_construct)))
        if(construct.circular):
            for pos,_ in enumerate(construct):
                cp_construct = construct.get_circularly_permuted(pos)
                
                if(cp_construct.get_species() == other_construct.get_species()):
                    return other_construct, (lambda a:((other_indexes[pos:]+other_indexes[:pos])[a],"f"))
                elif(cp_construct.get_reversed().get_species()==other_construct.get_species()):
                    return other_construct, (lambda a:((other_indexes[pos:]+other_indexes[:pos])[::-1][a],"r"))
        else:
            if(construct.get_species() == other_construct.get_species()):
                return other_construct, (lambda a: (a,"f"))
            elif(construct.get_reversed().get_species() == other_construct.get_species()):
                return other_construct, (lambda a: (other_indexes[::-1][a],"r"))

    def enumerate_components(self,components = None,previously_enumerated = None, **keywords):
        """this explores all the possible integrase-motivated DNA configurations. If some
        integrases aren't present, then define intnames to be a list of names of the
        integrases which are present.
        An integrase can act in different ways. 
        * serine integrases recombine B and P sites that turn into L and R sites, 
                    and only sites with the same dinucleotide can be recombined.
        * serine integrases with directionality factors recombine L and R sites
                    with the same dinucleotide
        * Invertases only do flipping reactions
        * resolvases only do deletion reactions
        * FLP or CRE react with homotypic sites, so site1+site1 = site1+site1. But
                    there are still different types of sites which are orthogonal. For
                    example, a CRE type 1 or a CRE type 2 site. The sites can also be palindromic,
                    which means that they can react in either direction.
        """
        if(previously_enumerated is None):
            previously_enumerated = []
        construct_list = []
        if(previously_enumerated is None):
            previously_enumerated = []
        for component in components:
            if(isinstance(component,DNA_construct)):
                construct_list += [component]
        int_dict = {}
        for construct in construct_list:
            #list each integrase that exists and which sites they react with
            
            con_dict = self.list_integrase(construct)
            
            int_dict = combine_dictionaries(int_dict,con_dict)
        constructlist = []
        for integrase in int_dict:
            if(integrase in self.int_mechanisms):
                int_mech = self.int_mechanisms[integrase]
                #now, going through each one, generate the reactions and species that arise
                attsites = int_dict[integrase]
                #but now we need to know what kind of integrase reactions are possible
                reactive_sites = int_mech.reactive_sites()
                attcombos = [a for a in it.combinations(attsites,2) if ((a[0].site_type in reactive_sites) and (a[1].site_type in reactive_sites))]
                for combo in attcombos:
                    #first question: is this combo legal?
                    if(int_mech.reaction_allowed(combo[0],combo[1])):
                        #this means the reaction can exist
                        #integrate now
                        new_dnas = int_mech.integrate(combo[0],combo[1],existing_dna_constructs = (previously_enumerated+constructlist))
                        constructlist += new_dnas
                
        return constructlist




# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

import libsbml
import warnings
def biocrnpyler_supported_units():
    supported_units = {
        # Volume units
        'nL': 
        {'unit_kind':[libsbml.UNIT_KIND_LITRE],
        'unit_exponents':1,
        'unit_scale':-9,
        'unit_multiplier':1},

        'uL': 
        {'unit_kind':libsbml.UNIT_KIND_LITRE,
        'unit_exponents':[1],
        'unit_scale':-6,
        'unit_multiplier':1},

        'mL': 
        {'unit_kind':libsbml.UNIT_KIND_LITRE,
        'unit_exponents':1,
        'unit_scale':[-3],
        'unit_multiplier':1},

        'L': 
        {'unit_kind':libsbml.UNIT_KIND_LITRE,
        'unit_exponents':1,
        'unit_scale':0,
        'unit_multiplier':[1]},
        
        # Concentration units
        'M': 
        {'unit_kind':[libsbml.UNIT_KIND_MOLE, libsbml.UNIT_KIND_LITRE],
        'unit_exponents':[1, -1],
        'unit_scale':[0, 0],
        'unit_multiplier':[1,1]},

        'mM': 
        {'unit_kind':[libsbml.UNIT_KIND_MOLE, libsbml.UNIT_KIND_LITRE],
        'unit_exponents':[1,-1],
        'unit_scale':[-3,0],
        'unit_multiplier':[1,1]},

        'uM': 
        {'unit_kind':[libsbml.UNIT_KIND_MOLE, libsbml.UNIT_KIND_LITRE],
        'unit_exponents':[1, -1],
        'unit_scale':[-6, 0],
        'unit_multiplier':[1,1]},

        'nM': 
        {'unit_kind':[libsbml.UNIT_KIND_MOLE, libsbml.UNIT_KIND_LITRE],
        'unit_exponents':[1, -1],
        'unit_scale':[-9, 0],
        'unit_multiplier':[1,1]},

        'mM': 
        {'unit_kind':[libsbml.UNIT_KIND_MOLE, libsbml.UNIT_KIND_LITRE],
        'unit_exponents':[1, -1],
        'unit_scale':[-3, 0],
        'unit_multiplier':[1,1]},
        
        # Time units
        'hour': 
        {'unit_kind':[libsbml.UNIT_KIND_SECOND],
        'unit_exponents':[1],
        'unit_scale':[0],
        'unit_multiplier':[3600]},

        'minute': 
        {'unit_kind':[libsbml.UNIT_KIND_SECOND],
        'unit_exponents':[1],
        'unit_scale':[0],
        'unit_multiplier':[60]},

        'second': 
        {'unit_kind':[libsbml.UNIT_KIND_SECOND],
        'unit_exponents':[1],
        'unit_scale':[0],
        'unit_multiplier':[1]},
        
        # Common parameter units
        'per_second': 
        {'unit_kind':[libsbml.UNIT_KIND_SECOND],
        'unit_exponents':[-1],
        'unit_scale':[0],
        'unit_multiplier':[1]},

        'per_minute': 
        {'unit_kind':[libsbml.UNIT_KIND_SECOND],
        'unit_exponents':[-1],
        'unit_scale':[0],
        'unit_multiplier':[1/60]},

        'per_hour': 
        {'unit_kind':[libsbml.UNIT_KIND_SECOND],
        'unit_exponents':[-1],
        'unit_scale':[0],
        'unit_multiplier':[1/3600]},

        'mole_per_litre': 
        {'unit_kind':[libsbml.UNIT_KIND_LITRE, libsbml.UNIT_KIND_MOLE],
        'unit_exponents':[-1,1],
        'unit_scale':[0,0],
        'unit_multiplier':[1,1]},

        'litre_per_mole_per_second': 
        {'unit_kind':[libsbml.UNIT_KIND_SECOND, libsbml.UNIT_KIND_LITRE, libsbml.UNIT_KIND_MOLE],
        'unit_exponents':[-1,1,-1],
        'unit_scale':[0,0,0],
        'unit_multiplier':[1,1,1]},

        'litre_per_mole_per_hour': 
        {'unit_kind':[libsbml.UNIT_KIND_SECOND, libsbml.UNIT_KIND_LITRE, libsbml.UNIT_KIND_MOLE],
        'unit_exponents':[-1,1,-1],
        'unit_scale':[0,0,0],
        'unit_multiplier':[1/3600,1,1]},
    }
    ### Add your own units to this dictionary ###
    return supported_units

def create_new_unit_definition(model, unit_id):
    """
    Creates a new UnitDefinition inside the 
    SBML Model object passed as model argument and other attributes. 
    Returns a pointer to the new libSBML object created for the unit type.
    """
    supported_units = biocrnpyler_supported_units()
    if type(unit_id) is not str:
        raise ValueError(
            'The arguments are not of expected type. unit_id must be a string.')
    if unit_id not in supported_units.keys():
        warnings.warn('The string identifier for the unit {0} is not supported by BioCRNpyler. Add this to the dictionary in biocrnpyler/units.py if you want this unit.'.format(unit_id))
        return None
    unit_kind = supported_units[unit_id]['unit_kind']
    unit_exponents = supported_units[unit_id]['unit_exponents']
    unit_scale = supported_units[unit_id]['unit_scale']
    unit_multiplier = supported_units[unit_id]['unit_multiplier']

    unitdef = model.createUnitDefinition()
    unitdef.setId(unit_id)
    # Scale list
    if type(unit_kind) is list and type(unit_scale) is not list:
        unit_scaleList = []
        for kind in unit_kind:
            unit_scaleList.append(unit_scale)

    elif type(unit_kind) is list and type(unit_scale) is list:
        if len(unit_kind) != len(unit_scale):
            raise ValueError(
                'Lengths of unit_scale and unit kind lists are not equal')
        unit_scaleList = unit_scale[:]

    # Multiplier list
    if type(unit_kind) is list and type(unit_multiplier) is not list:
        unit_multiplierList = []
        for kind in unit_kind:
            unit_multiplierList.append(unit_multiplier)

    elif type(unit_kind) is list and type(unit_multiplier) is list:
        if len(unit_kind) != len(unit_multiplier):
            raise ValueError(
                'Lengths of unit kind and unit_multiplier lists are not equal')
        unit_multiplierList = unit_multiplier[:]

    if type(unit_kind) is not list:
        unit_kind = [unit_kind]
        if type(unit_scale) is list:
            unit_scale = unit_scale[0]
        if type(unit_multiplier) is list:
            unit_multiplier = unit_multiplier[0]
        if type(unit_exponents) is list:
            unit_exponents = unit_exponents[0]
        if type(unit_scale) is not int or type(unit_multiplier) is not int:
            raise ValueError(
                'Scale and unit_multiplier must be integers when there is only one unit kind')
        unit_scaleList = [unit_scale]
        unit_multiplierList = [unit_multiplier]
    if type(unit_exponents) is not list:
        if type(unit_exponents) is not int:
            raise ValueError('All unit_exponentss should be integers')
        unit_exponents = [unit_exponents]
    if len(unit_kind) != len(unit_exponents):
        raise ValueError(
            'Lengths of unit kind and unit unit_exponents lists must be equal')

    for kind, expo, unit_scale, unit_multiplier in zip(unit_kind, unit_exponents, unit_scaleList, unit_multiplierList):
        unit = unitdef.createUnit()
        unit.setKind(kind)
        unit.setExponent(expo)
        unit.setScale(unit_scale)
        unit.setMultiplier(unit_multiplier)
    return unitdef



################################################################
#       DNA_construct: a higher level for construct compilation
#       Author: Andrey Shur
#       Latest update: 12/21/2020
#
################################################################


# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

import copy
from warnings import warn

from .dna_part_cds import CDS
from .dna_part_promoter import Promoter
from .dna_part_rbs import RBS
from .dna_part_terminator import Terminator
from .dna_construct import Construct, RNA_construct
from .component_enumerator import LocalComponentEnumerator
from .utils import all_comb, remove_bindloc, rev_dir
import logging


class ConstructExplorer(LocalComponentEnumerator):
    def __init__(self, name, direction="forward",possible_directions=("forward","reverse")):
        LocalComponentEnumerator.__init__(self=self,name=name)
        self.direction=direction
        self.possible_directions = possible_directions
        self.reset_enumerator()

    def reset_enumerator(self):
        """clear accumulator variables from the numerator. MUST BE SUBCLASSED"""
        pass

    def enumerate_components(self, component, previously_enumerated = None):
        #Only works on Constructs!
        self.reset_enumerator()
        if isinstance(component, Construct):
            #if we are circular, then we can go around twice
            #otherwise, once is the correct amount
            if(component.circular):
                self.max_loop_count = 2
            else:
                self.max_loop_count = 1
            #iterate through the possible directions
            for direction in self.possible_directions:
                #Set the loop count
                self.current_loop_count = 0 #due to while loop, current_loop_count is instantly increment to 1
                self.initialize_loop(direction = direction)

                #Deep Copy Component
                parts_list = list(component.parts_list)
                #Set the direction
                if direction == "reverse":
                    parts_list.reverse()
                #Main iteration loop
                while self.check_loop():
                    #Check each part
                    for part_ind, part in enumerate(parts_list):
                        #Do something to the part
                        self.iterate_part(part, direction)
                
                #must stop before switching directions
                self.terminate_loop()

            return self.return_components(component, previously_enumerated)
        else:
            return []

    def check_loop(self):
        """if we already went around the plasmid, then what we're checking for is continuing
        transcripts or proteins. We don't want to start making new transcripts
        because we already checked this area for promoters"""
        logging.debug(f"loop count = {self.current_loop_count}")
        if self.current_loop_count >= self.max_loop_count:
            self.terminate_loop()
            return False
        else:
            self.current_loop_count += 1
            return True

    def initialize_loop(self, direction):
        #Starts the loop
        #MUST SUBCLASS
        pass

    def iterate_part(self, part, direction):
        #Part is a DNA_part
        #direction is the direction we are looking in along the DNA_construct
        #runs on every part and does something!
        #MUST SUBCLASS
        pass

    def terminate_loop(self):
        #called when we are done enumerating
        pass
        #MUST SUBCLASS

    def return_components(self,component, previously_enumerated = None):
        #returns components at the end of the loop
        #MUST SUBCLASS
        return []

class TxExplorer(ConstructExplorer):
    def __init__(self, name = "TxExplorer", direction="forward",possible_directions=("forward","reverse")):
        ConstructExplorer.__init__(self, name, direction=direction,possible_directions=possible_directions)


    def reset_enumerator(self):
        self.all_rnas = {} #must clear this before filling it up again

    def initialize_loop(self, direction=None):
        self.current_rnas = {} #Stores RNA's being examined in loop promoter --> [(DNA_part, direction) list]


    def iterate_part(self, part, reading_direction):
        #Part is a DNA_part
        #direction is the direction we are looking in along the DNA_construct
        
        """the explorer sees a dna_part.
        1. it calculates effective direction of part relative to transcription direction
        2. it adds parts to growing transcripts
        3. it makes new transcripts and terminates growing transcripts.
        """
        logging.debug("looking at "+str(part))
        logging.debug("current_rnas are "+str(self.current_rnas))

        #The relative orientation of the part to the reading direction
        if part.direction == reading_direction:
            effective_direction = "forward"
        else:
            effective_direction = "reverse"

        #Grow Existing Transcripts
        #For each promoter being transcribed, we will add the part and its direction relative to that promoter
        for promoter in self.current_rnas:
            self.current_rnas[promoter]+= [(part, effective_direction)]

        #Case for Different Parts doing different things

        #Case 1: Promoter in the correct direction (and first loop around a circular plasmid)
        if effective_direction=="forward" and isinstance(part,Promoter) and self.current_loop_count == 1:
            #this if statement makes sure that the current part wants to transcribe, and
            #that is something that we are allowed to do
            self.current_rnas[part] = []

        #Case 2: Terminator
        elif effective_direction == "forward" and isinstance(part, Terminator): #Kill Sarah Day O'Connor
            self.terminate_loop()
    

    def terminate_loop(self):
        #Transfers current rnas into all_rnas

        #Create RNA_Constructs for each promoter
        for promoter,rna_parts_list in self.current_rnas.items():
            if(len(rna_parts_list) > 0):
                rna_construct = RNA_construct(rna_parts_list, promoter = promoter)
                self.all_rnas[promoter] = rna_construct
                promoter.transcript = rna_construct.get_species()
            else:
                warn(f"{promoter} makes empty transcript! Protein binding to this promoter will not be included in the CRN")
        self.initialize_loop()

    #Returns a list of RNAconstructs
    def return_components(self,component, previously_enumerated = None):
        return_rnas = []
        for rna in self.all_rnas.values():
            if((previously_enumerated is None) or (rna not in previously_enumerated)):
                return_rnas+=[rna]
            else:
                pass
        return return_rnas

class TlExplorer(ConstructExplorer):
    def __init__(self, name = "TlExplorer", direction="forward",possible_directions=("forward",)):
        ConstructExplorer.__init__(self, name, direction=direction,possible_directions=possible_directions)
    def reset_enumerator(self):
        self.all_proteins = {} #Stores all proteins from the RNA_construct
    def initialize_loop(self, direction=None):
        self.current_proteins = {} #Stores proteins being examined in loop rbs --> [(DNA_part, direction) list]
    def iterate_part(self, part, reading_direction):
        #Part is a DNA_part
        #direction is the direction we are looking in along the DNA_construct
        
        """the explorer sees a dna_part.
        1. it calculates effective direction of part relative to transcription direction
        2. it adds parts to growing transcripts
        3. it makes new transcripts and terminates growing transcripts.
        """
        logging.debug("looking at "+str(part))
        logging.debug("current_proteins are "+str(self.current_proteins))

        #The orientation of the part absolutely to the DNA_construct
        part_direction = part.direction
        #print("my part is ",part)
        #print("my direction is ",part.direction)

        #The relative orientation of the part to the reading direction
        if part_direction == reading_direction:
            effective_direction = "forward"
        else:
            effective_direction = "reverse"
        
        #count the number of RBSes that have been seen
        #currently this can only be 0 or 1 because RBS coupling is not implemented
        rbses_translating = len(self.current_proteins)

        #Case for Different Parts doing different things

        #Case 1: RBSes in the correct orientation
        if(isinstance(part,RBS) and effective_direction == "forward"):    
            if(rbses_translating>0):
                #rbs stops translation
                self.terminate_loop()
            #start a new protein
            self.current_proteins[part]= []

        #Case 2: CDSes in the correct orientation
        elif(rbses_translating >0 and isinstance(part,CDS) and effective_direction=="forward"):
            
            for rbs in self.current_proteins:
                #add to what is being translated
                self.current_proteins[rbs] += [[part,effective_direction]]
            #CDSes that aren't specifically designed to allow read-through
            #will terminate by default
            if(effective_direction not in part.no_stop_codons):
                #if the part has stop codons, terminate
                self.terminate_loop()

        #Case 3: all other parts usually terminate, unless they are designed to allow read-through
        elif(rbses_translating >0):
            if(effective_direction not in part.no_stop_codons):
                #if the part has stop codons, terminate
                self.terminate_loop()
                
    def terminate_loop(self):
        self.all_proteins.update(self.current_proteins)
        self.initialize_loop()

    def return_components(self,component, previously_enumerated = None):
        returnable_proteins = []
        for rbs in self.all_proteins:
            proteins = [a[0] for a in self.all_proteins[rbs]]
            returnable_proteins += proteins
            protein_species = [a.get_species() for a in proteins]
            if(protein_species==[]):
                protein_species = None
            component[rbs.position].protein = protein_species
        return returnable_proteins #these are CDS components



#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.


from .species import Complex, ComplexSpecies, OrderedPolymerSpecies, PolymerConformation
from .component import Component
from .components_combinatorial_complex import CombinatorialComplex
from . mechanisms_conformation import One_Step_Reversible_Conformation_Change
from .dna_part_promoter import Promoter
from itertools import permutations
import copy
import warnings

class CombinatorialConformation(Component):
    """
    A class to represent a PolymerConformation (made of one unique OrderedPolymerSpecies) with many internal Complexes which can bind and unbind in many different ways.
    """
    
    def __init__(self, final_states, initial_states = None, intermediate_states = None, excluded_states = None, state_part_ids = None, name = None, **keywords):
        """
        Binding reactions will be generated to form all PolymerConformations in final_states 
        from all the PolymerConformations in initial_states. There must be a single, unique, OrderedPolymerSpecies in all the conformations.
        Intermediate states restricts the binding reactions to first form PolymerConformations in this list. 
        At a high level this generates the following reactions:

        intial_states <-[Combinatorial Binding]-> final_states

        if  intermediate_states are given:
            intial_states <-[Combinatorial Binding]-> intermediate_states <-[Combinatorial Binding]->final_states
        
        
        Unlike CombinatorialComplex where Species are added individual, in CombinatorialConformation, groups of Species are added in single steps to produce the appropriate Complexes.

        :param final_states: one or more PolymerConformations.
        :param initial_states: a list of initial PolymerConformations which can bind/unbind to become the final_state
        :param intermediate_states: a list of intermediate PolymerConformations formed when converting initial_states to final_states. 
                                    If None, all possible intermediate PolymerConformations are enumerated.
        :param excluded_states: a list of intermediate PolymerConformations which will not be formed during enumeration.
                                if None: no intermediates will be excluded.
        :param state_part_ids: a dictionary {PolymerConformation : str} used to generate shorter part-ids for this conformation
        """

        if state_part_ids is None:
            self.state_part_ids = {}
        else:
            self.state_part_ids = state_part_ids
        self.internal_polymer = None #this will be set inside final_states setter
        self.final_states = final_states
        self.initial_states = initial_states
        self.intermediate_states = intermediate_states
        self.excluded_states = excluded_states

        if name is None:
            name = str(self.internal_polymer.name)
        Component.__init__(self, name = name, **keywords)
        self.add_mechanism(One_Step_Reversible_Conformation_Change())


    #Helper function to assert the correct class type
    def _assert_conformation(self, states, input_name = "states"):
        if not all([isinstance(s, PolymerConformation) for s in states]):
            raise ValueError(f"{input_name} must be a list of PolymerConformations. Recieved: {states}.")
        if not all([len(s.polymers) == 1 for s in states]):
            raise ValueError(f"All PolymerConformations in {input_name} must contain a single unique internal OrderedPolymerSpecies. Recieved: {states}.")

        if self.internal_polymer is None:
            self.internal_polymer = copy.deepcopy(states[0].polymers[0])
            self.internal_polymer.parent = None

        if not all([str(p) == str(self.internal_polymer) for s in states for p in s.polymers]):
            raise ValueError(f"All PolymerConformations in {input_name} must contain a single unique internal OrderedPolymerSpecies. Recieved: {states}.")

    def get_species(self):
        return PolymerConformation(polymer = self.internal_polymer)

    #Getters and setters
    #Final States stores the end complexes that will be formed
    @property
    def final_states(self):
        return self._final_states

    @final_states.setter
    def final_states(self, final_states):
        final_states = list(self.set_species(final_states))

        self._assert_conformation(final_states, "final_states")
        self._final_states = final_states

    #Initial states stores the starting states used in binding reactions
    @property
    def initial_states(self):
        return self._initial_states

    @initial_states.setter
    def initial_states(self, initial_states):
        #set initial states
        if initial_states is None or len(initial_states) == 0:
            self._initial_states = [PolymerConformation(polymer = self.internal_polymer)]
        else:
            initial_states = list(self.set_species(initial_states))
            #all initial_states must be PolymerConformation
            self._assert_conformation(initial_states, "initial_states")

            self._initial_states = initial_states

    #Intermediate states allows the user to restrict the complexes formed between the intial state and final state
    @property
    def intermediate_states(self):
        return self._intermediate_states
    @intermediate_states.setter
    def intermediate_states(self, intermediate_states):
        if intermediate_states is None:
            self._intermediate_states = None
        else:
            intermediate_states = list(self.set_species(intermediate_states))

            #All intermediate_states must be PolymerConformations
            self._assert_conformation(intermediate_states, "intermediate_states")

            self._intermediate_states = intermediate_states

    #excluded_states are PolymerConformations which are not allowed to form
    @property
    def excluded_states(self):
        return self._excluded_states
    @excluded_states.setter
    def excluded_states(self, excluded_states):
        if excluded_states is None:
            self._excluded_states = []
        else:
            #All excluded states must be PolymerConformations
            excluded_states = list(self.set_species(excluded_states))
            self._assert_conformation(excluded_states, "excluded_states")
            self._excluded_states = excluded_states

    def compute_species_changes(self, s0, sf):
        #print("computing species changes between", s0, "and", sf)
        #Compute Species that need to be added to s0 to get the PolymerConformation sf
        #Assumes the underlying internal polymer is the same
        #Computes a list of moves to go from s0 --> sf. Each move produces one of the Complexes in SF


        #1. if c0 contains bound locations not in cf, c0 cannot be transformed additively into cf
        if any([len(s0.get_complexes_at(0, i)) > len(sf.get_complexes_at(0, i)) for i in range(len(self.internal_polymer))]):
            species_changes = False
        else:
            species_changes = {}
            merged_complexes = {}

            #Figure out what happens to create each complex in sf
            for cf in sf.complexes:
                pf_inds = sf.get_polymer_positions(cf, 0)

                #initially, assume all external Species should be added
                #species are stored as (s, index) in order to differentiate species in different parts of the polymer
                cf_species_and_inds = [(cf.species[i], pf_inds[i])  for i in range(len(cf.species))]
                cf_species_and_inds_str = [str(i) for i in cf_species_and_inds] #string version used to check for equality
                sub_complex_species_and_inds = []
                sub_complex_species_and_inds_str = []
                merged_complexes[cf, pf_inds] = []


                #Iterate through speces in the polymer sf
                for ind in pf_inds:

                    #get complex of the polymer at this location
                    found_complexes = s0.get_complexes_at(0, ind)
                    for c0 in found_complexes:
                        p0_inds = s0.get_polymer_positions(c0, 0)
                        #Ensure each subcomplex c0 is only evaluated once
                        #then add all the Species in the complex to the sub_complex_species_and_inds list
                        if (cf, pf_inds) not in merged_complexes or c0 not in merged_complexes[cf, pf_inds]:
                            merged_complexes[cf, pf_inds].append(c0)

                            sub_complex_species_and_inds += [(c0.species[i], p0_inds[i])  for i in range(len(c0.species))]
                            sub_complex_species_and_inds_str += [str(i)  for i in sub_complex_species_and_inds]


                #cf cannot be created additively from the complexes in s0
                sub_complex_exclusive_species = [i for i in sub_complex_species_and_inds_str if i not in cf_species_and_inds_str]
                cf_complex_exclusive_species = [i for i in cf_species_and_inds_str if i not in sub_complex_species_and_inds_str]
                if len(sub_complex_exclusive_species) > 0:
                    species_changes[cf, pf_inds] = False
                #keep track of all the external species added to produce the new complex
                elif len(cf_complex_exclusive_species) > 0:
                    species_changes[cf, pf_inds] = [cf_species_and_inds[i][0] for i in range(len(cf_species_and_inds_str)) if (cf_species_and_inds_str[i] not in sub_complex_species_and_inds_str and cf_species_and_inds[i][1] is None)]
        #If any complex cannot be created, return False
        if species_changes == False or any([species_changes[k] is False for k in species_changes]) or (len(species_changes) == 0 and len(merged_complexes) == 0):
            #print("cannot convert.")
            return False
        else:
            #print("species_changes =", species_changes)
            #print("merged_complexes =", merged_complexes)
            return species_changes, merged_complexes



    def get_combinations_between(self, s0, sf):
        """
        Returns a list of 
        """
        #print("geting combinations between", s0, "and", sf)
        X = self.compute_species_changes(s0, sf)


        if X is False:
            return []
        else:
            species_changes, merged_complexes = X
            changes_list = []
            for cf in sf.complexes:
                inds = sf.get_polymer_positions(cf, 0)
                #print("checking ", cf, "inds =", inds, "in species_changes", (cf, inds) in species_changes, "in merged_complexes", (cf, inds) in merged_complexes)

                if ((cf, inds) in merged_complexes and len(merged_complexes[(cf, inds)]) == 1 and merged_complexes[(cf, inds)][0].monomer_eq(cf)) or (cf, inds) not in merged_complexes:
                    complexes_to_merge = None
                else:
                    complexes_to_merge = merged_complexes.get((cf, inds))

                if (cf, inds) not in species_changes or len(species_changes[cf, inds]) == 0:
                    species_to_add = None
                else:
                    species_to_add = species_changes.get((cf, inds))

                #print("(cf, inds, species_to_add, complexes_to_merge)=", (cf, inds, species_to_add, complexes_to_merge))
                if species_to_add is not None or complexes_to_merge is not None:
                    changes_list.append((cf, inds, species_to_add, complexes_to_merge))


            #get all permutations of the species for different binding orders 
            perms = permutations(changes_list, len(changes_list))

            #combinations (binder, bindee, complex_species) to be returned
            combinations = []

            #Iterate through permutations
            for perm in perms:
                old_state = s0
                for cf, inds, species_to_add, complexes_to_merge in perm:

                    #Determine all the species that are bound together
                    species_list = []
                    for i, p_loc in enumerate(inds):
                        if p_loc is None:
                            species_list.append(cf.species[i])
                        else:
                            species_list.append(old_state.polymers[0][p_loc])

                    new_complex = Complex(species_list)

                    #Determine which complexes to remove from the conformation
                    if complexes_to_merge is None:
                        merged_complexes = None
                    else:
                        merged_complexes = [new_complex.parent.get_complex(c) for c in complexes_to_merge]

                    #create a neew polymer conformation by removing the merged complexes
                    if merged_complexes not in [[], None]:
                        new_state = new_complex.parent.copy_remove_complexes(merged_complexes)
                    else:
                        new_state = new_complex.parent

                    #append the new combination if it isn't excluded
                    if old_state not in self.excluded_states and new_state not in self.excluded_states:
                        combinations.append((old_state, species_to_add, new_state))

                    #update bindee
                    old_state = new_state
            #print("combinations = ", combinations)
            return combinations

    def _get_part_id(self, state):
        if state in self.state_part_ids:
            return self.state_part_ids[state]
        else:
            return str(state)

    def update_species(self):
        mech_c = self.get_mechanism("conformation_change")
        species = []
        species_added_dict = {} #save which combinations have already been added
        self.combination_dict = {} #this should be recomputed every updated species

        #If there are intermediates, compute combinations in two steps
        if self.intermediate_states is not None:
            
            for s0 in self.initial_states:
                for si in self.intermediate_states:
                    if si != s0:
                        #Get combinatorial species between s0 and si
                        if (s0, si) not in self.combination_dict:
                            self.combination_dict[s0, si] = self.get_combinations_between(s0, si)

                        #iterate through combinations of species between s0 and si
                        for (s00, additional_species, sff) in self.combination_dict[s0, si]:
                            part_id = f"{self._get_part_id(s00)}-{self._get_part_id(sff)}"
                            species += mech_c.update_species(s0 = s00, sf = sff, additional_species = additional_species, 
                            component = self, part_id = part_id)

            for si in self.intermediate_states:
                for sf in self.final_states:
                    if si != sf:
                        #Get combinatorial species between si and sf
                        if (si, sf) not in self.combination_dict:
                            self.combination_dict[si, sf] = self.get_combinations_between(si, sf)

                        #iterate through combinations of species between si and sf
                        for (s00, additional_species, sff) in self.combination_dict[si, sf]:
                            part_id = f"{self._get_part_id(s00)}-{self._get_part_id(sff)}"
                            species += mech_c.update_species(s0 = s00, sf = sff, additional_species = additional_species, 
                            component = self, part_id = part_id)
            
        #If there are no intermediate restrictions, compute combinations in onestep
        else:
            for s0 in self.initial_states:
                for sf in self.final_states:
                    if s0 != sf:
                        #Get combinatorial species between s0 and sf
                        if (s0, sf) not in self.combination_dict:
                            self.combination_dict[s0, sf] = self.get_combinations_between(s0, sf)
                            
                        #iterate through combinations of species between s0 and sf
                        for (s00, additional_species, sff) in self.combination_dict[s0, sf]:
                            part_id = f"{self._get_part_id(s00)}-{self._get_part_id(sff)}"
                            species += mech_c.update_species(s0 = s00, sf = sff, additional_species = additional_species, 
                            component = self, part_id = part_id)

        return list(set(species))

    def update_reactions(self):
        mech_c = self.get_mechanism("conformation_change")
        reactions = []
        reactions_added_dict = {} #save which combinations have already been added in order to not add duplicates

        #If there are intermediates, compute combinations in two steps
        if self.intermediate_states is not None:
            for s0 in self.initial_states:
                for si in self.intermediate_states:
                    #Get combinatorial species between s0 and si
                    if (s0, si) not in self.combination_dict:
                        self.combination_dict[s0, si] = self.get_combinations_between(s0, si)

                    #iterate through combinations of species between s0 and si
                    for (s00, additional_species, sff) in self.combination_dict[s0, si]:
                        if (s00, sff) not in reactions_added_dict:
                            reactions_added_dict[s00, sff] = True
                            part_id = f"{self._get_part_id(s00)}-{self._get_part_id(sff)}"
                            reactions += mech_c.update_reactions(s0 = s00, sf = sff, additional_species = additional_species, 
                            component = self, part_id = part_id)

            for si in self.intermediate_states:
                for sf in self.final_states:
                    #Get combinatorial species between si and sf
                    if (si, sf) not in self.combination_dict:
                        self.combination_dict[si, sf] = self.get_combinations_between(si, sf)

                    #iterate through combinations of species between si and sf
                    for (s00, additional_species, sff) in self.combination_dict[si, sf]:
                        if (s00, sff) not in reactions_added_dict:
                            reactions_added_dict[s00, sff] = True
                            part_id = f"{self._get_part_id(s00)}-{self._get_part_id(sff)}"
                            reactions += mech_c.update_reactions(s0 = s00, sf = sff, additional_species = additional_species, 
                            component = self, part_id = part_id)

        #If there are no intermediate restrictions, compute combinations in onestep
        else:
            for s0 in self.initial_states:
                for sf in self.final_states:
                    #Get combinatorial species between s0 and sf
                    if (s0, sf) not in self.combination_dict:
                        self.combination_dict[s0, sf] = self.get_combinations_between(s0, sf)
                        
                    #iterate through combinations of species between s0 and sf
                    for (s00, additional_species, sff) in self.combination_dict[s0, sf]:
                        if (s00, sff) not in reactions_added_dict:
                            reactions_added_dict[s00, sff] = True
                            part_id = f"{self._get_part_id(s00)}-{self._get_part_id(sff)}"
                            reactions += mech_c.update_reactions(s0 = s00, sf = sff, additional_species = additional_species, 
                            component = self, part_id = part_id)

        return reactions



class CombinatorialConformationPromoter(CombinatorialConformation, Promoter):
    """
    A combinatorial conformaiton with an additional set of states "expressing_states" which can transcribe/express rna/protein products.
    This class merges Promoter and CombinatorialConformation.

        :param promoter_states: one or more PolymerConformations which are used by the promoter class
        :param promoter_states_on: True/False if True all promoter_states are transcribable. If False all states except promoter_states are transcribable
        :param promoter_location: the index of the monomer in the PolymerConformation which represents the promoter
        :param final_states: one or more PolymerConformations.
        :param initial_states: a list of initial PolymerConformations which can bind/unbind to become the final_state
        :param intermediate_states: a list of intermediate PolymerConformations formed when converting initial_states to final_states. 
                                    If None, all possible intermediate PolymerConformations are enumerated.
        :param excluded_states: a list of intermediate PolymerConformations which will not be formed during enumeration.
                                if None: no intermediates will be excluded.
        :param state_part_ids: a dictionary {PolymerConformation : str} used to generate shorter part-ids for this conformation
        :param activating_complexes: a list of ComplexSpecies which activate PolymerConformations allowing them to be transcribed.
        :param inactivating_complexes: a list of ComplexSpecies which innactive the PolymerConformation, preventing transcription.

    """
    def __init__(self, promoter_states, promoter_location, promoter_states_on = True, activating_complexes = None, inactivating_complexes = None,
        name = "CombinatorialConformationPromoter", **keywords):
        
        Promoter.__init__(self, name, **keywords)
        CombinatorialConformation.__init__(self, name = name, **keywords)

        self.promoter_states = promoter_states
        self.promoter_states_on = promoter_states_on

        if activating_complexes is None:
            self.activating_complexes = []
        else:
            self.activating_complexes = activating_complexes

        if inactivating_complexes is None:
            self.inactivating_complexes = []
        else:
            self.inactivating_complexes = inactivating_complexes

        assert all([isinstance(c, ComplexSpecies) for c in self.activating_complexes])
        assert all([isinstance(c, ComplexSpecies) for c in self.inactivating_complexes])

        if promoter_location not in range(len(self.internal_polymer)):
            raise ValueError(f"promoter_location must be an index of the polymer {self.internal_polymer}. Recieved {promoter_location}.")
        else:
            self.promoter_location = promoter_location


    #promoter_states are PolymerConformations which are used by the Promoter class
    #These can be ON or OFF
    @property
    def promoter_states(self):
        return self._promoter_states

    @promoter_states.setter
    def promoter_states(self, promoter_states):
        if promoter_states is None:
            self._promoter_states = []
        else:
            #All excluded states must be PolymerConformations
            promoter_states = list(self.set_species(promoter_states))
            self._assert_conformation(promoter_states, "promoter_states")
            self._promoter_states = promoter_states


    def update_species(self):
        self.conformation_species = CombinatorialConformation.update_species(self)
        promoter_species = []
        old_name = self.name
        for s in self.conformation_species:
            if isinstance(s, PolymerConformation):
                active_state = False
                if  (s in self.promoter_states and self.promoter_states_on) or (s not in self.promoter_states and not self.promoter_states_on):
                    active_state = True

                active_complex = False
                if any([s.get_complex(c) is not None for c in self.activating_complexes]):
                    active_complex = True

                innactive_complex = False
                if any([s.get_complex(c) is not None for c in self.inactivating_complexes]):
                    innactive_complex = True
                    if (active_state and len(self.promoter_states)>0) or active_complex:
                        warnings.warn("Inactive_complex conflicts with active_complex or active_state in CombinatorialConformationPromoter. Defaulting to inactive.")
                        
                if (active_state or active_complex) and not innactive_complex:
                    self.dna_to_bind = s.polymers[0][self.promoter_location]
                    self.name = self._get_part_id(s) #Reset name for unique addressable promoter states
                    promoter_species += Promoter.update_species(self)
        
        return list(set(promoter_species+self.conformation_species))

    def update_reactions(self):
        if not hasattr(self, "conformation_species"):
            self.update_species

        conformation_reactions = CombinatorialConformation.update_reactions(self)
        promoter_reactions = []
        old_name = self.name
        for s in self.conformation_species:
            if isinstance(s, PolymerConformation):
                active_state = False
                if  (s in self.promoter_states and self.promoter_states_on) or (s not in self.promoter_states and not self.promoter_states_on):
                    active_state = True

                active_complex = False
                if any([s.get_complex(c) is not None for c in self.activating_complexes]):
                    active_complex = True

                innactive_complex = False
                if any([s.get_complex(c) is not None for c in self.inactivating_complexes]):
                    innactive_complex = True
                    if (active_state and len(self.promoter_states)>0) or active_complex:
                        warnings.warn("innactive_complex conflicts with active_complex or active_state in CombinatorialConformationPromoter. Defaulting to innactive.")
                        
                if (active_state or active_complex) and not innactive_complex:
                    self.dna_to_bind = s.polymers[0][self.promoter_location]
                    self.name = self._get_part_id(s) #Reset name for unique addressable promoter states
                    promoter_reactions += Promoter.update_reactions(self)
        self.name = old_name
        return promoter_reactions + conformation_reactions



#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.
from typing import List, Union
from warnings import warn

class ComponentEnumerator:
    def __init__(self, name:str):
        """a component enumerator's job is to create new components in a process similar to mechanisms."""
        self.name = name

    def enumerate_components(self,component=None, **keywords) -> List:
        """this will create new components based on the input component somehow
        The child class should implement this
        :return: empty list"""
        warn(f"Default update_components called for ComponentEnumerator = {self.name}.")
        return []

    def __repr__(self):
        return self.name

class LocalComponentEnumerator(ComponentEnumerator):
    """a component enumerator's job is to create new components in a process similar to mechanisms.
    A local component enumerator only cares about the single component that is passed in"""
    def __init__(self,name:str):
        ComponentEnumerator.__init__(self,name=name)

class GlobalComponentEnumerator(ComponentEnumerator):
    def __init__(self,name:str):
        """a component enumerator's job is to create new components in a process similar to mechanisms.
    A global component enumerator takes in every component that is in the mixture. This is for complex enumeration that
    cares about other components"""
        ComponentEnumerator.__init__(self,name=name)
        



#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

from typing import Dict, List
from warnings import warn

from .mechanism import Mechanism
from .mechanisms_enzyme import MichaelisMenten
from .reaction import Reaction
from .species import ComplexSpecies, OrderedPolymerSpecies, Species

"""
Global mechanisms are a lot like mechanisms. They are called only by mixtures
 on a list of all species have been generated by components. Global mechanisms are meant
 to work as universal mechanisms that function on each species or all species of some
 type or with some attribute. Global mechanisms may only act on one species at a time.

In order to decide which species a global mechanism acts upon, the filter_dict is used.
      filter_dict[species.material_type / specie.attributes / species.name / repr(species)] = True / False
 For each species, its material type, name, and attributes are sent through the filter_dict. If True
 is returned, the GlobalMechanism will act on the species. If False is returned, the
 the GlobalMechanism will not be called. If there are conflicts in the filter_dict, an error is raised.

If the species's name, material type, and attributes are all not in the filter_dict, the GlobalMechanism will
 be called if default_on == True and not called if default_on == False.
Note that the above filtering is done automatically. Any parameters needed by the global mechanism must be
 in the Mixture's parameter dictionary. These methods are assumed to take a single species
 as input.

An example of a global mechanism is degradation via dilution which is demonstrated in the Tests folder.

GlobalMechanisms should be used cautiously or avoided alltogether - the order in which they are called
may have to be carefully user defined in the subclasses of Mixture in order to ensure expected behavior.
"""


class GlobalMechanism(Mechanism):
    """Global mechanisms are a lot like mechanisms. They are called only by mixtures
    on a list of all species have been generated by components. Global mechanisms
    are meant to work as universal mechanisms that function on each species or
    all species of some material type or with some attribute. Global mechanisms
    may only act on one species at a time.

    In order to decide which species a global mechanism acts upon, the filter_dict
    is used.

    An example of a global mechanism is degradation via dilution which is
    demonstrated in the Tests folder.

    GlobalMechanisms should be used cautiously or avoided alltogether - the order
    in which they are called may have to be carefully user-defined in the
    subclasses of Mixture in order to ensure expected behavior.
    """
    def __init__(self, name: str, mechanism_type: str="", filter_dict: Dict=None,
                 default_on: bool=False, recursive_species_filtering: bool=False):
        """Creates a GlobalMechanisms instance.

        If the species's name, material type, and attributes are all not in the
        filter_dict, the GlobalMechanism will be called if default_on == True and
        not called if default_on == False.

        :param name: name of the GlobalMechanism
        :param mechanism_type:
        :param filter_dict: filter_dict[species.material_type / species.attributes] = True / False
         For each species, its material type or attributes are sent through the
        filter_dict. If True is returned, the GlobalMechanism will act on the
        species. If False is returned, the the GlobalMechanism will not be called.
        If filter_dict[attribute] is different from filter_dict[material_type],
        filter_dict[attribute] takes precedent. If multiple filter_dict[attribute]
        contradict for different attributes, an error is raised.
        Note that the above filtering is done automatically. Any parameters needed by
        the global mechanism must be in the Mixture's parameter dictionary. These
        methods are assumed to take a single species as input.
        :param default_on: what to do if a species doesn't come up in the filter dict. Also used for as the default if there is a filterdict conflict
        :param recursive_species_filtering:  keyword determines how the material_type and name of ComplexSpecies is defined.
        If True: the filter based upon all subspecies.type and name recursively going through
        all ComplexSpecies. If False: the filter dict will act only on the ComplexSpecies. By default, this is False.
        """
        if filter_dict is None:
            self.filter_dict = {}
        else:
            self.filter_dict = filter_dict

        self.default_on = default_on
        self.recursive_species_filtering = recursive_species_filtering
        Mechanism.__init__(self, name=name, mechanism_type=mechanism_type)

    def apply_filter(self, s: Species):
        """applies the filter dictionary to determine if a global mechanism acts on a species.

        :param s: Species
        :return:
        """
        fd = self.filter_dict
        use_mechanism = None
        species_list = s.get_species(recursive=self.recursive_species_filtering)
        for subs in species_list:
            for a in subs.attributes+[subs.material_type, subs.name]:
                if a in fd:
                    if use_mechanism is None:
                        use_mechanism = fd[a]
                    elif use_mechanism != fd[a]:
                        warn(f"species {repr(s)} has multiple attributes(or material type) which conflict with global mechanism filter {repr(self)}. Using default value {self.default_on}.")
                        use_mechanism = self.default_on

        if use_mechanism is None:
            use_mechanism = self.default_on
        return use_mechanism

    def update_species_global(self, species_list: List[Species], mixture):
        new_species = []
        for s in species_list:
            use_mechanism = self.apply_filter(s)
            if use_mechanism:
                new_species += self.update_species(s, mixture)
        return new_species

    def update_reactions_global(self, species_list: List[Species], mixture):
        fd = self.filter_dict
        new_reactions = []
        for s in species_list:
            use_mechanism = self.apply_filter(s)
            if use_mechanism:
                new_reactions += self.update_reactions(s, mixture)

        return new_reactions

    def get_parameter(self, species, param_name, mixture):
        param = mixture.get_parameter(mechanism = self, part_id = repr(species), param_name = param_name)
        if param is None:
            raise ValueError("No parameters can be found that match the "
                 "(mechanism, part_id, "
                f"param_name)=({repr(self)}, {repr(species)}, "
                f"{param_name}).")
        else:
            return param

    def update_species(self, s: Species, mixture):
        """All global mechanisms must use update_species functions with these inputs.

        :param s: Species instance
        :return:
        """
        return []

    def update_reactions(self, s, mixture):
        """All global mechanisms must use update_reactions functions with these inputs.

        :param s:
        :param mixture:
        :return:
        """
        return []


class Dilution(GlobalMechanism):
    """A global mechanism to represent dilution."""

    def __init__(self, name = "global_degredation_via_dilution",
                 mechanism_type = "dilution", filter_dict=None,
                 default_on = True, recursive_species_filtering = True):
        GlobalMechanism.__init__(self, name = name,
                                 mechanism_type = mechanism_type,
                                 default_on = default_on,
                                 filter_dict = filter_dict,
                                 recursive_species_filtering = recursive_species_filtering)

    def update_reactions(self, s: Species, mixture):
        k_dil = self.get_parameter(s, "kdil", mixture)
        rxn = Reaction.from_massaction(inputs=[s], outputs=[], k_forward=k_dil)
        return [rxn]


class AnitDilutionConstiutiveCreation(GlobalMechanism):
    """Global Mechanism to Constitutively Create Certain Species at the rate of dilution.

    Useful for keeping machinery species like ribosomes and polymerases at a constant concentration.
    """
    def __init__(self, name = "anti_dilution_constiuitive_creation",
                 material_type="dilution", filter_dict=None,
                 default_on = True,  recursive_species_filtering = True):
        GlobalMechanism.__init__(self, name = name,
                                 mechanism_type = material_type,
                                 default_on = default_on,
                                 filter_dict = filter_dict, 
                                 recursive_species_filtering = recursive_species_filtering)

    def update_reactions(self, s, mixture):
        k_dil = self.get_parameter(s, "kdil", mixture)
        rxn = Reaction.from_massaction(inputs=[], outputs=[s], k_forward=k_dil)
        return [rxn]



class Degredation_mRNA_MM(GlobalMechanism, MichaelisMenten):
    """Michaelis Menten mRNA Degredation by Endonucleases
       mRNA + Endo <--> mRNA:Endo --> Endo
       All species of type "rna" are degraded by this mechanisms, including those inside of a ComplexSpecies.
       ComplexSpecies are seperated by this process, including embedded ComplexSpecies. 
       OrderedPolymerSpecies are ignored.
    """
    def __init__(self, nuclease, name="rna_degredation_mm", mechanism_type = "rna_degredation", 
        default_on = False, recursive_species_filtering = True, filter_dict = None, **keywords):

        if isinstance(nuclease, Species):
            self.nuclease = nuclease
        else:
            raise ValueError("'nuclease' must be a Species.")
        MichaelisMenten.__init__(self=self, name=name, mechanism_type = mechanism_type)

        if filter_dict is None:
            filter_dict = {"rna":True, "notdegradable":False}

        GlobalMechanism.__init__(self, name = name, mechanism_type = mechanism_type, default_on = default_on,
                                 filter_dict = filter_dict, recursive_species_filtering = recursive_species_filtering)

    def update_species(self, s, mixture):
        species = []

        #Check if rna species are inside a ComplexSpecies. 
        #If so, break up the ComplexSpecies and degrade the RNA
        if isinstance(s, ComplexSpecies) and s.material_type != "rna" and not isinstance(s, OrderedPolymerSpecies):
            internal_species = s.get_species(recursive = True)
            non_rna_species = [sp for sp in internal_species if sp.material_type != "rna" and sp != s]
            if len(non_rna_species)>0:
                prod = non_rna_species
            else:
                prod = None
            species += MichaelisMenten.update_species(self, Enzyme = self.nuclease, Sub = s, Prod = prod)

        #If the material type is simply RNA, break it up.
        elif s.material_type == "rna":
            species += MichaelisMenten.update_species(self, Enzyme = self.nuclease, Sub = s, Prod = None)
        else:
            #This case includes OrderedPolymerSpecies with RNA inside them and species with RNA in their name (but not mateiral type)
            species = []

        return species

    def update_reactions(self, s, mixture):
        reactions = []

        #Check if rna species are inside a ComplexSpecies. 
        #If so, break up the ComplexSpecies and degrade the RNA

        

        if isinstance(s, ComplexSpecies) and s.material_type != "rna" and not isinstance(s, OrderedPolymerSpecies):
            kdeg = self.get_parameter(s, "kdeg", mixture)
            kb = self.get_parameter(s, "kb", mixture)
            ku = self.get_parameter(s, "ku", mixture)

            internal_species = s.get_species(recursive = True)
            non_rna_species = [sp for sp in internal_species if sp.material_type != "rna" and sp != s]
            if len(non_rna_species)>0:
                prod = non_rna_species
            else:
                prod = None
            reactions += MichaelisMenten.update_reactions(self, Enzyme = self.nuclease, Sub = s, Prod = prod, kb=kb, ku=ku, kcat=kdeg)

        #If the material type is simply RNA, break it up.
        elif s.material_type == "rna":
            kdeg = self.get_parameter(s, "kdeg", mixture)
            kb = self.get_parameter(s, "kb", mixture)
            ku = self.get_parameter(s, "ku", mixture)
            
            reactions += MichaelisMenten.update_reactions(self, Enzyme = self.nuclease, Sub = s, Prod = None, kb=kb, ku=ku, kcat=kdeg)
        else:
            #This case includes OrderedPolymerSpecies with RNA inside them and species with RNA in their name (but not mateiral type)
            reactions = []

        return reactions


class Deg_Tagged_Degredation(GlobalMechanism, MichaelisMenten):
    """Michaelis Menten Degredation of deg-tagged proteins by degredase (such as proteases)
       Species_degtagged + degredase <--> Species_degtagged:degredase --> degredase
       All species with the attribute degtagged and material_type protein are degraded. The method is not recursive.
    """
    def __init__(self, degredase, deg_tag = "degtagged", name="deg_tagged_degredation", mechanism_type="degredation", 
        filter_dict= None, recursive_species_filtering = False, default_on = False, **keywords):
        if isinstance(degredase, Species):
            self.degredase = degredase
        else:
            raise ValueError("'degredase' must be a Species.")
        MichaelisMenten.__init__(self=self, name=name, mechanism_type=mechanism_type)

        if filter_dict is None:
            filter_dict = {deg_tag:True}
            
        GlobalMechanism.__init__(self, name = name, mechanism_type = mechanism_type, default_on = default_on,
                                 filter_dict = filter_dict, recursive_species_filtering = recursive_species_filtering)

    def update_species(self, s, mixture):
        species = []
        species += MichaelisMenten.update_species(self, Enzyme = self.degredase, Sub = s, Prod = None)
        return species

    def update_reactions(self, s, mixture):

        kdeg = self.get_parameter(s, "kdeg", mixture)
        kb = self.get_parameter(s, "kb", mixture)
        ku = self.get_parameter(s, "ku", mixture)

        rxns = []
        rxns += MichaelisMenten.update_reactions(self, Enzyme = self.degredase, Sub = s, Prod = None, kb=kb, ku=ku, kcat=kdeg)
        return rxns



# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

import warnings

class Compartment():

    """ A formal Compartment object for a Species in a CRN.
     A Compartment must have a name. They may also have a spatial dimension (such as 2 for two-dimensional,
     or 3 for three-dimensional) and the size in litres. 
     Note: The "default" keyword is reserved for BioCRNpyler allotting a default compartment. 
     Users must choose a different string. 
     The unit attribute for Compartment can be used to set unit for the Compartment size. Make sure
     that the string identifier used for the unit is a supported unit in BioCRNpyler. Check the 
     documentation to find a list of supported units. Add your own custom units in units.py if needed.
    """

    def __init__(self, name: str, size = 1e-6, spatial_dimensions = 3, unit = None, **keywords):
        self.name = name
        self.spatial_dimensions = spatial_dimensions 
        self.size = size 
        self.unit = unit 

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):
        if name is None:
            raise TypeError("Compartment name must be a string.")
        elif type(name) is str:
            no_underscore_string = name.replace("_", "")
            if no_underscore_string.isalnum() and "__" not in name and name[len(name)-1] != "_" and name[0].isalpha():
                self._name = name
            else:
                raise ValueError(f"name attribute {name} must consist of letters, numbers, or underscores and cannot contained double underscores or begin/end with a special character.")
        else:
            raise ValueError('Compartment name must be a string.')

    @property
    def spatial_dimensions(self):
        return self._spatial_dimensions

    @spatial_dimensions.setter
    def spatial_dimensions(self, spatial_dimensions: int):
        if type(spatial_dimensions) is not int:
            raise ValueError('Compartment spatial dimension must be an integer.')
        elif spatial_dimensions < 0:
            raise ValueError('Compartment spatial dimension must be non-negative.')
        else:
            self._spatial_dimensions = spatial_dimensions

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, size: float):
        if type(size) not in [float, int]:
            raise ValueError('Compartment size must be a float or int.')
        elif size < 0:
            raise ValueError('Compartment size must be non-negative.')
        else:
            self._size = size
    
    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, unit: str):
        if unit is not None:
            if not isinstance(unit, str):
                raise ValueError("Unit of compartment must be a string representing compartment size.")
            self._unit = unit 
        else:
            self._unit = None

    def __eq__(self, other):
        """
        Overrides the default implementation
        Two compartments are equivalent if they have the same name, spatial dimension, and size
        :param other: Compartment instance
        :return: boolean
        """
        if isinstance(other, Compartment) and self.name == other.name:
            # Now if two compartments have same name but other attributes are different, throw an error:
            if self.size != other.size or self.spatial_dimensions != other.spatial_dimensions:
                raise ValueError('Compartments with same names must have the same size and spatial dimensions.')
            return True
        else:
            return False



#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.


from .species import Complex, ComplexSpecies
from .component import Component
from itertools import permutations
import copy

class CombinatorialComplex(Component):
    """
    A class to represent a Complex of many Species which can bind together in many different ways.
    """
    
    def __init__(self, final_states, initial_states = None, intermediate_states = None, excluded_states = None, name = None, **keywords):
        """
        Binding reactions will be generated to form all the ComplexSpecies in final_states from all the species in initial_states 
        (or, if initial_states is None, from all the individual species inside each ComplexSpecies). Intermediate states restricts
        the binding reactions to only form species in this list. Excluded states are not allowed to be reactants or products. 
        At a high level this generates the following reactions:
        
        If just final_states are given:
            final_states_internal_species <-[Combinatorial Binding]-> final_states
        
        if initial_states are given:
            intial_states <-[Combinatorial Binding]-> final_states

        if intermediate_states are given:
            final_states_internal_species <-[Combinatorial Binding]-> intermediate_states <-[Combinatorial Binding]->final_states

        if initial_states and intermediate_states are given:
            intial_states <-[Combinatorial Binding]-> intermediate_states <-[Combinatorial Binding]->final_states
        

        :param final_states: a single ComplexSpecies or a list of ComplexSpecies. 
        :param initial_states: a list of initial Species which are bound together to form the ComplexSpecies in final_states. 
                                If None defaults to the members of the ComplexSpecies in final_states.
        :param intermediate_states: a list of intermediate ComplexSpecies formed when converting initial_states to final_states. 
                                    If None: all possible intermediate ComplexSpecies are enumerated.
        :param excluded_states: a list of ComplexSpecies which are NOT allowed to form when converting initial states to final states.
                                If None: no ComplexSpecies are excluded.


        Example 1: final_states = ComplexSpecies([A, B, C]). initial_states = None, intermediate_states = None.
                   initial_states will default to A, B, C. All intermediate states [A, B], [A, C], [B, C] will be enumerated. 
                   This results in the 6 reversible reactions:
                   1. A + B <--> Complex([A, B])
                   2. A + C <--> Complex([A, C])
                   3. B + C <--> Complex([B, C])
                   4. Complex([A, B]) + C <--> Complex([A, B, C])
                   5. Complex([A, C]) + B <--> Complex([A, B, C])
                   6. Complex([B, C]) + A <--> Complex([A, B, C])

        Example 2: final_states = ComplexSpecies([A, B, C]), intial_states = [Complex([A, B]), Complex([A, C])], intermediate_states = None
                   This results in the reactions:
                   1. Complex([A, B]) + C <--> Complex([A, B, C])
                   2. Complex([A, C]) + B <--> Complex([A, B, C])
        
        Example 3: final_states = ComplexSpecies([A, B, C]). initial_states = None, intermaied_states = [Complex([A, B]), Complex([A, C])].
                   This results in reactions:
                   1. A + B <--> Complex([A, B])
                   2. A + C <--> Complex([A, C])
                   3. Complex([A, B]) + C <--> Complex([A, B, C])
                   4. Complex([A, C]) + B <--> Complex([A, B, C])

        Example 4: final_states = [Complex([A, A, B], Complex([A, B, B]))], initial_states = None, intermediate_states = None
                   This results in the reactions:
                   1. A + A <--> Complex([A, A])
                   2. Complex([A, A]) + B <--> Complex ([A, A, B])
                   3. B + B <--> Complex([B, B])
                   4. Complex([B, B]) + A <--> Complex ([A, B, B])
                   5. A + B <--> Complex([A, B])
                   6. Complex([A, B]) + A <--> Complex ([A, A, B])
                   7. Complex([A, B]) + B <--> Complex ([A, B, B])
        """

        #The order these run in is important!

        #1. set final_states
        self.final_states = final_states
        #2. set initial_states
        self.initial_states = initial_states
        #3. set intermidiate_states
        self.intermediate_states = intermediate_states
        #4. set excluded_states
        self.excluded_states = excluded_states

        #used to store combinations of species during update
        self.combination_dict = {}

        #Call super
        if name is None:
            name = ""
            for s in self.final_states:
                name += s.name+"_"
            name = name[:-1]
        super().__init__(name, **keywords)

    #Final States stores the end complexes that will be formed
    @property
    def final_states(self):
        return self._final_states

    @final_states.setter
    def final_states(self, final_states):
        final_states = list(self.set_species(final_states))

        #all final_states must be ComplexSpecies
        if not all([isinstance(s, ComplexSpecies) for s in final_states]):
            raise ValueError(f"final_states must be a list of {ComplexSpecies} (or subclasses thereof). Recieved: {final_states}.")

        self._final_states = final_states

        #Then create a list of all sub-species included in final_states Complexes
        self.sub_species = []
        for s in self.final_states:
            self.sub_species += s.species_set

    #Initial states stores the starting states used in binding reactions
    @property
    def initial_states(self):
        return self._initial_states

    @initial_states.setter
    def initial_states(self, initial_states):
        #set initial states
        if initial_states is None:
            self._initial_states = self.sub_species
        else:
            initial_states = list(self.set_species(initial_states))
            for s in initial_states:
                if not (s in self.sub_species or (isinstance(s, ComplexSpecies) and all([ss in self.sub_species for ss in s.species_set]))):
                    raise ValueError(f"Invalid initial species {s}; initial_states must either be contained in the final_states or a {ComplexSpecies} made of Species in the final_states.")
            self._initial_states = initial_states

    #Intermediate states allows the user to restrict the complexes formed between the intial state and final state
    @property
    def intermediate_states(self):
        return self._intermediate_states
    @intermediate_states.setter
    def intermediate_states(self, intermediate_states):
        if intermediate_states is None:
            self._intermediate_states = None
        else:
            intermediate_states = list(self.set_species(intermediate_states))

            #All intermediate_states must be ComplexSpecies or OrderdedComplexSpecies
            if not all([isinstance(s, ComplexSpecies) for s in intermediate_states]):
                raise ValueError(f"intermediate must be a list of {ComplexSpecies} (or subclasses thereof). Recieved: {intermediate_states}.")
            #All intermediate_states must be made of sub_species
            for s in intermediate_states:
                intermediate_sub_species = s.species_set
                if not all([ss in self.sub_species for ss in intermediate_sub_species]):
                    raise ValueError(f"intermediate species {s} contains subspecies not in the final_states.")

            self._intermediate_states = intermediate_states

    #Excluded states allows the user to exclude specific Species from being enumerated
    @property
    def excluded_states(self):
        return self._excluded_states
    @excluded_states.setter
    def excluded_states(self, excluded_states):
        if excluded_states is None:
            self._excluded_states = []
        else:
            self._excluded_states = list(self.set_species(excluded_states))

    def compute_species_to_add(self, s0, sf):
        #Compute Species that need to be added to s0 to get the Complex sf

        if not isinstance(sf, ComplexSpecies):
            raise ValueError(f"sf must be a ComplexSpecies. Recieved {sf}")

        species_to_add = []
        for s in sf.species_set:

            if s0.monomer_eq(s): #this is used instead of == to deal with the potential for different parents
                s0_count = 1
            elif isinstance(s0, ComplexSpecies):
                s0_count = s0.monomer_count(s)#This is used instead of .count to deal with species with different parents
            else:
                s0_count = 0
            
            #Add the correct number of s to the list
            sf_count = sf.monomer_count(s)
            if s0_count < sf_count:
                species_to_add += (sf_count-s0_count)*[s]
            elif s0_count > sf_count:
                species_to_add = None #In this case, s0 contains more stuff than sf, so nothing should be returned
                break
            else:
                pass #if they have the same number, do not add it

        #s0 contains more or different species than sf, return None
        if (not isinstance(s0, ComplexSpecies)) and (sf.monomer_count(s0) == 0):
            species_to_add = None
        elif isinstance(s0, ComplexSpecies) and any([s0.monomer_count(s) > sf.monomer_count(s) for s in s0.species_set]) and sf.monomer_count(s0) == 0:
            species_to_add = None
        elif len(species_to_add) == 0:
            species_to_add = None

        return species_to_add

    def get_combinations_between(self, s0, sf):
        """
        Returns all combinations of Species to create the Complex sf from s0.
        """

        species_to_add = self.compute_species_to_add(s0, sf)
        
        if species_to_add is None or len(species_to_add) == 0:
            return []
        else:
            #combinations (binder, bindee, complex_species) to be returned
            combinations = []
            #get all permutations of the species for different binding orders 
            perms = permutations(species_to_add, len(species_to_add))

            #Iterate through permutations
            for perm in perms:
                bindee = s0
                for s in perm:
                    binder = s
                    if isinstance(bindee, ComplexSpecies):
                        s_list = list(bindee.species)
                    else:
                        s_list = [bindee]

                    s_list.append(s)
                    cs = Complex(s_list)

                    #append the new combination if it isn't excluded
                    if binder not in self.excluded_states and bindee not in self.excluded_states and cs not in self.excluded_states:
                        combinations.append((binder, bindee, cs))
                    #update bindee
                    bindee = cs

            return combinations
            
    def update_species(self):
        mech_b = self.get_mechanism("binding")
        species = []
        species_added_dict = {} #save which combinations have already been added
        self.combination_dict = {} #this should be recomputed every updated species

        #If there are intermediates, compute combinations in two steps
        if self.intermediate_states is not None:
            for s0 in self.initial_states:
                for si in self.intermediate_states:
                    #Get combinatorial species between s0 and si
                    if (s0, si) not in self.combination_dict:
                        self.combination_dict[s0, si] = self.get_combinations_between(s0, si)

                    #iterate through combinations of species between s0 and si
                    for binder, bindee, cs in self.combination_dict[s0, si]:
                        if (binder, bindee, cs) not in species_added_dict:
                            species_added_dict[binder, bindee, cs] = True
                            species += mech_b.update_species(binder = binder, bindee = bindee, complex_species = cs, 
                            component = self, part_id = self.name+"_"+str(binder)+"_"+str(bindee))

            for si in self.intermediate_states:
                for sf in self.final_states:
                    #Get combinatorial species between si and sf
                    if (si, sf) not in self.combination_dict:
                        self.combination_dict[si, sf] = self.get_combinations_between(si, sf)

                    #iterate through combinations of species between si and sf
                    for binder, bindee, cs in self.combination_dict[si, sf]:
                        if (binder, bindee, cs) not in species_added_dict:
                            species_added_dict[binder, bindee, cs] = True
                            species += mech_b.update_species(binder = binder, bindee = bindee, complex_species = cs, 
                            component = self, part_id = self.name+"_"+str(binder)+"_"+str(bindee))

        #If there are no intermediate restrictions, compute combinations in onestep
        else:
            for s0 in self.initial_states:
                for sf in self.final_states:
                    #Get combinatorial species between s0 and sf
                    if (s0, sf) not in self.combination_dict:
                        self.combination_dict[s0, sf] = self.get_combinations_between(s0, sf)
                        
                    #iterate through combinations of species between s0 and sf
                    for binder, bindee, cs in self.combination_dict[s0, sf]:
                        if (binder, bindee, cs) not in species_added_dict:
                            species_added_dict[binder, bindee, cs] = True
                            species += mech_b.update_species(binder = binder, bindee = bindee, complex_species = cs, 
                            component = self, part_id = self.name+"_"+str(binder)+"_"+str(bindee))

        return list(set(species))

    def update_reactions(self):
        mech_b = self.get_mechanism("binding")
        reactions = []
        reactions_added_dict = {} #save which combinations have already been added in order to not add duplicates

        #If there are intermediates, compute combinations in two steps
        if self.intermediate_states is not None:
            for s0 in self.initial_states:
                for si in self.intermediate_states:
                    #Get combinatorial species between s0 and si
                    if (s0, si) not in self.combination_dict:
                        self.combination_dict[s0, si] = self.get_combinations_between(s0, si)

                    #iterate through combinations of species between s0 and si
                    for binder, bindee, cs in self.combination_dict[s0, si]:
                        if (binder, bindee, cs) not in reactions_added_dict and (bindee, binder, cs) not in reactions_added_dict:
                            reactions_added_dict[binder, bindee, cs] = True
                            reactions += mech_b.update_reactions(binder = binder, bindee = bindee, complex_species = cs, 
                            component = self, part_id = self.name+"_"+str(binder)+"_"+str(bindee))

            for si in self.intermediate_states:
                for sf in self.final_states:
                    #Get combinatorial species between si and sf
                    if (si, sf) not in self.combination_dict:
                        self.combination_dict[si, sf] = self.get_combinations_between(si, sf)

                    #iterate through combinations of species between si and sf
                    for binder, bindee, cs in self.combination_dict[si, sf]:
                        if (binder, bindee, cs) not in reactions_added_dict and (bindee, binder, cs) not in reactions_added_dict:
                            reactions_added_dict[binder, bindee, cs] = True
                            reactions += mech_b.update_reactions(binder = binder, bindee = bindee, complex_species = cs, 
                            component = self, part_id = self.name+"_"+str(binder)+"_"+str(bindee))

        #If there are no intermediate restrictions, compute combinations in onestep
        else:
            for s0 in self.initial_states:
                for sf in self.final_states:
                    #Get combinatorial species between s0 and sf
                    if (s0, sf) not in self.combination_dict:
                        self.combination_dict[s0, sf] = self.get_combinations_between(s0, sf)
                        
                    #iterate through combinations of species between s0 and sf
                    for binder, bindee, cs in self.combination_dict[s0, sf]:
                        if (binder, bindee, cs) not in reactions_added_dict and (bindee, binder, cs) not in reactions_added_dict:
                            reactions_added_dict[binder, bindee, cs] = True
                            reactions += mech_b.update_reactions(binder = binder, bindee = bindee, complex_species = cs, 
                            component = self, part_id = self.name+"_"+str(binder)+"_"+str(bindee))

        return reactions





################################################################
#       DNA_part: a component-like intermediate class necessary for DNA_construct
#       Author: Andrey Shur
#       Latest update: 6/4/2020
#
#
################################################################


# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

from .component import Component
from .polymer import OrderedMonomer, OrderedPolymer
from .species import Species
import copy


class DNA_part(Component, OrderedMonomer):
    def __init__(self,name, **keywords):
        """this represents a modular component sequence. These get compiled into working components"""

        if "initial_concentration" in keywords:
            raise AttributeError("DNA_part should not recieve initial_concentration keyword. Pass this into the DNAassembly or DNA_construct instead.")

        Component.__init__(self=self, name = name, **keywords)
        
        self.name = name
        assembly = None #this is already covered in self.assembly. Is this needed?
        direction = None #orientation
        pos = None #position in the dna_construct
        self.sequence = None #nucleotide sequence
        #most parts have stop codons
        #if you want your part to not have stop codons, put "forward" and/or "reverse"
        self.no_stop_codons = [] #some parts will set this, others won't. Default is that it has stop codons
        self.material_type = "part"
        for key, value in keywords.items():
            #goes through any extra parameters and sets them!
            if(key=="assembly"):
                assembly = value
            elif(key=="direction"):
                direction = value
            elif(key=="pos"):
                pos = value
            elif(key=="sequence"):
                self.sequence = value
            elif(key=="no_stop_codons"):
                if(value is not None):
                    self.no_stop_codons = value
            elif(key=="material_type"):
                self.material_type = value
        if(isinstance(assembly,OrderedPolymer)):
            OrderedMonomer.__init__(self,position=pos,parent=assembly,direction=direction)
        else:
            self.assembly = assembly
            OrderedMonomer.__init__(self,position=pos,direction=direction)

    @property
    def dna_species(self):
        return Species(self.name, material_type="part")

    def __repr__(self):
        myname = self.name
        if(self.position is not None):
            myname+="_"+str(self.position)
        if(self.direction =="reverse"):
            myname += "_r"
        return myname
    def __hash__(self):
        return OrderedMonomer.__hash__(self)+hash(self.name)
    def __eq__(self,other):
        if(type(other)==type(self)):
            if(self.name==other.name):
                if(self.assembly is not None and other.assembly is not None):
                    if(str(self.assembly)==str(other.assembly)):
                        return True
                elif((self.assembly is not None) or (other.assembly is not None)):
                    #if one has an assembly and the other doesn't, then they aren't the same!!
                    return False
                elif(((self.parent is None) and (other.parent is None)) or \
                    ((self.parent is not None) and (other.parent is not None) and (str(self.parent)==str(other.parent)))):
                    #this is for when we are using the OrderedMonomer for its intended function
                    if(self.direction==other.direction and self.position==other.position):
                        return True
        return False

    def clone(self,position,direction,parent_dna):
        """this defines where the part is in what piece of DNA"""
        #TODO add warning if DNA_part is not cloned
        self.insert(parent_dna,position,direction)
        return self

    def unclone(self):
        """removes the current part from anything"""
        self.remove()
        return self
        
    def reverse(self):
        OrderedMonomer.reverse(self)
        return self



from .mechanism import Mechanism
from .mechanisms_enzyme import MichaelisMentenCopy
from .propensities import ProportionalHillNegative, ProportionalHillPositive
from .reaction import Reaction
from .species import Complex, Species
from .utils import parameter_to_value
from typing import List

class OneStepGeneExpression(Mechanism):
    """A mechanism to model gene expression without transcription or translation.

    G --> G + P
    """
    def __init__(self, name="gene_expression",
                 mechanism_type="transcription"):
        """Initializes a OneStepGeneExpression instance.

        :param name: name of the Mechanism, default: gene_expression
        :param mechanism_type: type of the Mechanism, default: transcription

        """
        Mechanism.__init__(self, name=name, mechanism_type=mechanism_type)

    def update_species(self, dna, protein, transcript=None, **keywords):
        species = [dna]
        if protein is not None:
            species += [protein]

        return species

    def update_reactions(self, dna, component = None, kexpress = None,
                         protein=None, transcript = None, part_id = None, **keywords):

        if kexpress is None and component is not None:
            kexpress = component.get_parameter("kexpress", part_id = part_id, mechanism = self)
        elif component is None and kexpress is None:
            raise ValueError("Must pass in component or a value for kexpress")

        if protein is not None:
            return [Reaction.from_massaction(inputs=[dna], outputs=[dna, protein], k_forward=kexpress)]
        else:
            return []


class SimpleTranscription(Mechanism):
    """A Mechanism to model simple catalytic transcription.

    G --> G + T
    """
    def __init__(self, name="simple_transcription", mechanism_type="transcription"):
        """Initializes a SimpleTranscription instance.

        :param name: name of the Mechanism, default: simple_transcription
        :param mechanism_type: type of the Mechanism, default: transcription

        """
        Mechanism.__init__(self, name=name, mechanism_type=mechanism_type)

    def update_species(self, dna, transcript = None, protein = None, **keywords):

        species = [dna]
        if transcript is not None:
            species += [transcript]
        if protein is not None:
            species += [protein]

        return species

    def update_reactions(self, dna, component = None, ktx = None, part_id = None, transcript = None, protein = None, **keywords):

        if ktx == None and component != None:
            ktx = component.get_parameter("ktx", part_id = part_id, mechanism = self)
        elif component == None and ktx == None:
            raise ValueError("Must pass in component or a value for ktx")

        #First case only true in Mixtures without transcription (eg Expression Mixtures)
        if transcript is None and protein is not None:
            ktl = component.get_parameter("ktl", part_id = part_id, mechanism = self)
            rxns = [Reaction.from_massaction(inputs = [dna], outputs = [dna, protein], k_forward=ktx * ktl)]
        else:
            rxns = [Reaction.from_massaction(inputs = [dna], outputs = [dna, transcript], k_forward=ktx)]

        return rxns


class SimpleTranslation(Mechanism):
    """A mechanism to model simple catalytic translation.

    T --> T + P
    """
    def __init__(self, name="simple_translation", mechanism_type="translation"):
        """Initializes a SimpleTranslation instance.

        :param name: name of the Mechanism, default: simple_translation
        :param mechanism_type: type of the Mechanism, default: translation

        """
        Mechanism.__init__(self, name=name, mechanism_type=mechanism_type)

    def update_species(self, transcript, protein = None,  **keywords):
        if protein is None:
            protein = Species(transcript.name, material_type="protein")
        outlst = [transcript]
        if(type(protein)==list):
            outlst+=protein
        else:
            outlst+=[protein]
        return outlst

    def update_reactions(self, transcript, component = None, ktl = None, part_id = None, protein = None, **keywords):

        if ktl is None and component is not None:
            ktl = component.get_parameter("ktl", part_id = part_id, mechanism = self)
        elif component is None and ktl is None:
            raise ValueError("Must pass in component or a value for ktl")

        # First case only true in Mixtures without transcription (eg Expression Mixtures)
        if transcript is None and protein is not None:
            rxns = []
        else:
            rxns = [Reaction.from_massaction(inputs = [transcript], outputs = [transcript, protein], k_forward=ktl)]

        return rxns


class PositiveHillTranscription(Mechanism):
    """A mechanism to model transcription as a proprotional positive hill function:
    G --> G + P
    rate = k*G*(R^n)/(K+R^n)
    where R is a regulator (activator).
    Optionally includes a leak reaction
    G --> G + P @ rate kleak.
    """

    def __init__(self, name="positivehill_transcription", mechanism_type="transcription"):
        """Initializes a PositiveHillTranscription instance.

        :param name: name of the Mechanism, default: positivehill_transcription
        :param mechanism_type: type of the Mechanism, default: transcription

        """
        Mechanism.__init__(self, name=name, mechanism_type=mechanism_type)

    def update_species(self, dna, regulator, transcript = None, leak = False, protein = None, **keywords):

        species = [dna, regulator]
        if transcript is not None:
            species += [transcript]
        if protein is not None:
            species += [protein]

        return species  # it is best to return all species that will be involved in the reactions

    def update_reactions(self, dna, regulator, component, part_id, transcript = None, leak = False, protein = None, **keywords):
        """This always requires the inputs component and part_id to find the relevant parameters

        :param dna:
        :param regulator:
        :param component:
        :param part_id:
        :param transcript:
        :param leak:
        :param protein:
        :param keywords:
        :return:
        """

        ktx = component.get_parameter("k", part_id = part_id, mechanism = self)
        n = component.get_parameter("n", part_id = part_id, mechanism = self)
        K = component.get_parameter("K", part_id = part_id, mechanism = self)
        kleak = component.get_parameter("kleak", part_id = part_id, mechanism = self)

        prophill = ProportionalHillPositive(k=ktx, K=K, s1=regulator, n=n, d=dna)

        reactions = []

        #First case only true in Mixtures without transcription (eg Expression Mixtures)
        if transcript is None and protein is not None:
            tx_output = protein
        else:
            tx_output = transcript

        reactions.append(Reaction(inputs=[dna], outputs=[dna, tx_output], propensity_type=prophill))

        if leak:
            reactions.append(Reaction.from_massaction(inputs=[dna], outputs=[dna, tx_output], k_forward=kleak))

        #In this case, we just return one reaction
        return reactions


class NegativeHillTranscription(Mechanism):
    """A mechanism to model transcription as a proprotional negative hill function:
    G --> G + P
    rate = k*G*(1)/(K+R^n)
    where R is a regulator (repressor).
    Optionally includes a leak reaction
    G --> G + P @ rate kleak.
    """

    def __init__(self, name="negativehill_transcription", mechanism_type="transcription"):
        """Initializes a NegativeHillTranscription instance.

        :param name: name of the Mechanism, default: negativehill_transcription
        :param mechanism_type: type of the Mechanism, default: transcription

        """
        Mechanism.__init__(self, name=name, mechanism_type=mechanism_type)

    def update_species(self, dna, regulator, transcript = None, leak = False, protein = None, **keywords):

        species = [dna, regulator]
        if transcript is not None:
            species += [transcript]
        if protein is not None:
            species += [protein]

        return species  # it is best to return all species that will be involved in the reactions

    def update_reactions(self, dna, regulator, component, part_id, transcript = None, leak = False, protein = None, **keywords):
        """This always requires the inputs component and part_id to find the relevant parameters

        :param dna:
        :param regulator:
        :param component:
        :param part_id:
        :param transcript:
        :param leak:
        :param protein:
        :param keywords:
        :return:
        """

        ktx = component.get_parameter("k", part_id = part_id, mechanism = self)
        n = component.get_parameter("n", part_id = part_id, mechanism = self)
        K = component.get_parameter("K", part_id = part_id, mechanism = self)
        kleak = component.get_parameter("kleak", part_id = part_id, mechanism = self)

        prop_hill = ProportionalHillNegative(k=ktx, K=K, n=n, s1=regulator, d=dna)

        reactions = []

        #First case only true in Mixtures without transcription (eg Expression Mixtures)
        if transcript is None and protein is not None:
            tx_output = protein
        else:
            tx_output = transcript

        reactions.append(Reaction(inputs=[dna], outputs=[dna, tx_output], propensity_type=prop_hill))

        if leak:
            reactions.append(Reaction.from_massaction(inputs = [dna], outputs = [dna, tx_output], k_forward=kleak))

        #In this case, we just return one reaction
        return reactions


class Transcription_MM(MichaelisMentenCopy):
    """Michaelis Menten Transcription.

        G + RNAP <--> G:RNAP --> G+RNAP+mRNA
    """

    def __init__(self, rnap: Species, name="transcription_mm", **keywords):
        """Initializes a Transcription_MM instance.

        :param rnap: Species instance that is representing an RNA polymerase
        :param name: name of the Mechanism, default: transcription_mm
        """
        if isinstance(rnap, Species):
            self.rnap = rnap
        else:
            raise ValueError("'rnap' parameter must be a Species.")

        MichaelisMentenCopy.__init__(self=self, name=name,
                                     mechanism_type="transcription")

    def update_species(self, dna, transcript=None, protein = None, **keywords):
        species = [dna]

        if transcript is None and protein is not None:
            tx_output = protein
        else:
            tx_output = transcript

        species += MichaelisMentenCopy.update_species(self, Enzyme = self.rnap, Sub = dna, Prod = tx_output)

        return species

    def update_reactions(self, dna, component, part_id = None, complex=None, transcript=None, protein = None,
                         **keywords):

        #Get Parameters
        if part_id == None and component != None:
            part_id = component.name

        ktx = component.get_parameter("ktx", part_id = part_id, mechanism = self)
        kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
        ku = component.get_parameter("ku", part_id = part_id, mechanism = self)

        rxns = []

        if transcript is None and protein is not None:
            tx_output = protein
        else:
            tx_output = transcript

        rxns += MichaelisMentenCopy.update_reactions(self, Enzyme = self.rnap, Sub = dna, Prod = tx_output, complex=complex, kb=kb, ku=ku, kcat=ktx)

        return rxns


class Translation_MM(MichaelisMentenCopy):
    """Michaelis Menten Translation.

        mRNA + Rib <--> mRNA:Rib --> mRNA + Rib + Protein
    """

    def __init__(self, ribosome: Species, name="translation_mm", **keywords):
        """Initializes a Translation_MM instance.

        :param ribosome: Species instance that is representing a ribosome
        :param name: name of the Mechanism, default: translation_mm
        """
        if isinstance(ribosome, Species):
            self.ribosome = ribosome
        else:
            raise ValueError("ribosome must be a Species!")
        MichaelisMentenCopy.__init__(self=self, name=name,
                                     mechanism_type="translation")

    def update_species(self, transcript, protein, **keywords):
        species = []

        #This can only occur in expression mixtures
        if transcript is None and protein is not None:
            species += Species.flatten_list([protein])
        else:
            species += MichaelisMentenCopy.update_species(self, Enzyme = self.ribosome, Sub = transcript, Prod = protein)

        return species

    def update_reactions(self, transcript, protein, component, part_id = None, complex=None, **keywords):
        rxns = []

        #Get Parameters
        if part_id == None and component != None:
            part_id = component.name

        ktl = component.get_parameter("ktl", part_id = part_id, mechanism = self)
        kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
        ku = component.get_parameter("ku", part_id = part_id, mechanism = self)


        #This can only occur in expression mixtures
        if transcript is None and protein is not None:
            pass
        else:
            rxns += MichaelisMentenCopy.update_reactions(self, Enzyme = self.ribosome, Sub = transcript, Prod = protein, complex=complex, kb=kb, ku=ku, kcat=ktl)
        return rxns

class Energy_Transcription_MM(Mechanism):
    """Michaelis Menten Transcription that consumed energy.

        G + RNAP <--> G:RNAP
        Fuel + G:RNAP --> G + RNAP + T + Fuel (Transcription can only happen when there is fuel)
            at rate ktx/L (length dependent transcription rate)

        Fuel + G:RNAP --> G:RNAP +wastes (Fuel consumption treated faster)
            at rate ktx (This occurs L times faster than the above, resulting in the correct fuel use)
    """

    def __init__(self, rnap: Species,  fuels: List[Species], wastes = List[Species], name="energy_transcription_mm", **keywords):
        """Initializes a Transcription_MM instance.
        :param fuels: List of Species consumed during transcription
        :param wastes: List of Species consumed during transcription
        :param rnap: Species instance that is representing an RNA polymerase
        :param name: name of the Mechanism, default: transcription_mm
        """
        if isinstance(rnap, Species):
            self.rnap = rnap
        else:
            raise ValueError("'rnap' parameter must be a Species.")

        if all([isinstance(s, Species) for s in fuels]):
            self.fuels = fuels
        else:
            raise ValueError('recieved a non-Species object in the fuels list.')

        if all([isinstance(s, Species) for s in fuels]):
            self.wastes = wastes
        else:
            raise ValueError("wastes must be a list of Species!")

        Mechanism.__init__(self=self, name=name,
                                     mechanism_type="transcription")

    def update_species(self, dna, transcript=None, protein = None, **keywords):
        species = [dna, self.rnap, transcript]+self.fuels
        bound_complex = Complex([dna, self.rnap])
        species += [bound_complex]

        return species

    def update_reactions(self, dna, component, part_id = None, complex=None, transcript=None, protein = None,
                         **keywords):

        #Get Parameters
        if part_id == None and component != None:
            part_id = component.name

        ktx = component.get_parameter("ktx", part_id = part_id, mechanism = self)
        kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
        ku = component.get_parameter("ku", part_id = part_id, mechanism = self)
        L = component.get_parameter("length", part_id = part_id, mechanism = self)


        bound_complex = Complex([dna, self.rnap])

        #RNAP DNA Binding
        r1 = Reaction.from_massaction([dna, self.rnap], [bound_complex], k_forward = kb, k_reverse = ku)
        #Transcription
        r2 = Reaction.from_massaction(self.fuels + [bound_complex], self.fuels + [dna, self.rnap, transcript],
            k_forward = parameter_to_value(ktx.value)/parameter_to_value(L))
        #Fuel consumption
        r3 = Reaction.from_massaction(self.fuels + [bound_complex], [bound_complex]+self.wastes, k_forward = ktx)

        return [r1, r2, r3]


class Energy_Translation_MM(Mechanism):
    """Michaelis Menten Translation that consumes energy species.

        mRNA + Rib  <--> mRNA:Rib  (binding)
        fuels + mRNA:Rib --> mRNA + Rib + Protein + fuels    (translation)
        fuels + mRNA:Rib --> mRNA:Rib  +wastes (fuel consumption)
    """

    def __init__(self, ribosome: Species, fuels: List[Species], wastes = List[Species], name="energy_translation_mm", **keywords):
        """Initializes a Translation_MM instance.
        
        :param ribosome: Species instance that is representing a ribosome
        :param fuels: List of fuel Species that are consumed during translation
        :param wastes: List of Species consumed during translation
        :param name: name of the Mechanism, default: energy_translation_mm
        """
        if isinstance(ribosome, Species):
            self.ribosome = ribosome
        else:
            raise ValueError("ribosome must be a Species!")

        if all([isinstance(s, Species) for s in fuels]):
            self.fuels = fuels
        else:
            raise ValueError("Fuels must be a list of Species!")

        if all([isinstance(s, Species) for s in fuels]):
            self.wastes = wastes
        else:
            raise ValueError("wastes must be a list of Species!")

        Mechanism.__init__(self=self, name=name,
                                     mechanism_type="translation")

    def update_species(self, transcript, protein, **keywords):
        species = self.fuels+[self.ribosome, protein]
        bound_complex = Complex([transcript, self.ribosome])
        species += [bound_complex]

        return species

    def update_reactions(self, transcript, protein, component, part_id = None, complex=None, **keywords):
        rxns = []

        #Get Parameters
        if part_id == None and component != None:
            part_id = component.name

        ktl = component.get_parameter("ktl", part_id = part_id, mechanism = self)
        kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
        ku = component.get_parameter("ku", part_id = part_id, mechanism = self)
        L = component.get_parameter("length", part_id = part_id, mechanism = self)


        bound_complex = Complex([transcript, self.ribosome])

        #RNAP DNA Binding
        r1 = Reaction.from_massaction([transcript, self.ribosome], [bound_complex], k_forward = kb, k_reverse = ku)
        #Transcription
        r2 = Reaction.from_massaction(self.fuels + [bound_complex], self.fuels + [transcript, self.ribosome, protein],
            k_forward = parameter_to_value(ktl.value)/parameter_to_value(L))
        #Fuel consumption
        r3 = Reaction.from_massaction(self.fuels + [bound_complex], [bound_complex]+self.wastes, k_forward = ktl)

        return [r1, r2, r3]

class multi_tx(Mechanism):
    """Multi-RNAp Transcription w/ Isomerization.

    Detailed transcription mechanism accounting for each individual
    RNAp occupancy states of gene.

    n ={0, max_occ}
    DNA:RNAp_n + RNAp <--> DNA:RNAp_n_c --> DNA:RNAp_n+1
    DNA:RNAp_n --> DNA:RNAp_0 + n RNAp + n mRNA
    DNA:RNAp_n_c --> DNA:RNAp_0_c + n RNAp + n mRNA

    n --> number of open configuration RNAp on DNA
    max_occ --> Physical maximum number of RNAp on DNA (based on RNAp and DNA dimensions)
    DNA:RNAp_n --> DNA with n open configuration RNAp on it
    DNA:RNAp_n_c --> DNA with n open configuration RNAp and 1 closed configuration RNAp on it

    For more details, see examples/MultiTX_Demo.ipynb
    """

    def __init__(self, pol: Species, name: str='multi_tx', mechanism_type: str='transcription', **keywords):
        """Initializes a multi_tx instance.

        :param pol: reference to a species instance that represents a polymerase
        :param name: name of the Mechanism, default: multi_tx
        :param mechanism_type: type of the mechanism, default: transcription
        :param keywords:
        """
        if isinstance(pol, Species):
            self.pol = pol
        else:
            raise ValueError("'pol' must be a Species")

        Mechanism.__init__(self, name=name, mechanism_type=mechanism_type)

    # species update
    def update_species(self, dna, transcript, component, part_id, protein = None, **keywords):
        max_occ = int(component.get_parameter("max_occ", part_id = part_id, mechanism = self, return_numerical = True))
        cp_open = []
        cp_closed = []
        for n in range(0,max_occ):
            cp_open.append(Complex([dna]+[self.pol for i in range(n+1)], attributes = ["open"]))
            cp_closed.append(Complex([dna]+[self.pol for i in range(n+1)], attributes = ["closed"]))

        cp_misc = [self.pol,dna,transcript]

        return cp_open + cp_closed + cp_misc

    def update_reactions(self, dna, transcript, component, part_id, protein = None, **keywords):
        """It sets up the following reactions.

        DNA:RNAp_n + RNAp <--> DNA:RNAp_n_c --> DNA:RNAp_n+1
        kf1 = k1, kr1 = k2, kf2 = k_iso
        DNA:RNAp_n --> DNA:RNAp_0 + n RNAp + n mRNA
        kf = ktx_solo
        DNA:RNAp_n_c --> DNA:RNAp_0_c + n RNAp + n mRNA
        kf = ktx_solo

        max_occ =  maximum occupancy of gene (physical limit)
        """

        # parameter loading
        kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
        ku = component.get_parameter("ku", part_id = part_id, mechanism = self)
        k_iso = component.get_parameter("k_iso", part_id = part_id, mechanism = self)
        ktx = component.get_parameter("ktx", part_id = part_id, mechanism = self)
        max_occ = int(component.get_parameter("max_occ", part_id = part_id, mechanism = self, return_numerical = True))

        # complex species instantiation
        cp_open = []
        cp_closed = []
        for n in range(0,max_occ):
            cp_open.append(Complex([dna]+[self.pol for i in range(n+1)], attributes = ["open"])) #has n polymerases all open
            cp_closed.append(Complex([dna]+[self.pol for i in range(n+1)], attributes = ["closed"])) #has n-1 open polymerases and 1 closed polymerase


        # Reactions
        # polymerase + complex(n) <--> complex(n+1)_closed
        rxn_open_p = [Reaction.from_massaction(inputs=[self.pol, cp_open[n]], outputs=[cp_closed[n+1]], k_forward=kb, k_reverse = ku) for n in range(0, max_occ-1)]
        #rxn_open_pr = [Reaction.from_massaction(inputs=[cp_closed[n + 1]], outputs=[self.pol, cp_open[n], ], k_forward=k2) for n in range(0, max_occ - 1)]

        # isomerization
        #complex(n)_closes --> complex(n)
        rxn_iso = [Reaction.from_massaction(inputs=[cp_closed[n]], outputs=[cp_open[n]], k_forward=k_iso) for n in range(0, max_occ)]

        # release/transcription from open and closed states
        rxn_release_open =  []
        rxn_release_closed = []
        for n in range(0,max_occ):
            rxn_temp1 = Reaction.from_massaction(inputs= [cp_open[n]], outputs=[self.pol for i in range(n+1)] +
                                                                               [transcript for i in range(n+1)] + [dna], k_forward=ktx)
            rxn_release_open.append(rxn_temp1)

        for n in range(1,max_occ):
            rxn_temp2 = Reaction.from_massaction(inputs= [cp_closed[n]], outputs=[self.pol for i in range(n)] +
                                                                                 [transcript for i in range(n)] + [cp_closed[0]], k_forward=ktx)
            rxn_release_closed.append(rxn_temp2)

        # base case pol + dna <--> complex(n=1)_open
        rxn_m1 = Reaction.from_massaction(inputs=[dna, self.pol], outputs=[cp_closed[0]], k_forward=kb, k_reverse = ku)

        rxn_all = rxn_open_p + rxn_iso + rxn_release_open + rxn_release_closed + [rxn_m1]

        return rxn_all


class multi_tl(Mechanism):
    """Multi-RBZ Translation w/ Isomerization.

    Detailed translation mechanism accounting for each individual
    RBZ occupancy states of mRNA. Still needs some work, so use with caution,
    read all warnings and consult the example notebook.

    n ={0, max_occ}
    mRNA:RBZ_n + RBZ <--> mRNA:RBZ_n_c --> mRNA:RBZ_n+1
    mRNA:RBZ_n --> mRNA:RBZ_0 + n RBZ + n Protein
    mRNA:RBZ_n_c --> mRNA:RBZ_0_c + n RBZ + n Protein

    n --> number of open configuration RBZ on mRNA
    max_occ --> Physical maximum number of RBZ on mRNA (based on RBZ and mRNA dimensions)
    mRNA:RBZ_n --> mRNA with n open configuration RBZ on it
    mRNA:RBZ_n_c --> mRNA with n open configuration RBZ and 1 closed configuration RBZ on it

    For more details, see examples/MultiTX_Demo.ipynb
    """

    def __init__(self, ribosome: Species, name: str='multi_tl', mechanism_type: str='translation', **keywords):
        """Initializes a multi_tl instance.

        :param ribosome: a Species instance that represents a ribosome
        :param name: name of the Mechanism, default: multi_tl
        :param mechanism_type: type of the Mechanism, default: translation

        """
        if isinstance(ribosome, Species):
            self.ribosome = ribosome
        else:
            raise ValueError("'ribosome' must be a Species.")

        Mechanism.__init__(self, name=name, mechanism_type=mechanism_type)

    # species update
    def update_species(self, transcript, protein, component, part_id, **keywords):
        max_occ = int(component.get_parameter("max_occ", part_id = part_id, mechanism = self, return_numerical = True))
        cp_open = []
        cp_closed = []
        for n in range(0, max_occ):
            cp_open.append(Complex([transcript]+[self.ribosome for i in range(n+1)], attributes = ["open"]))
            cp_closed.append(Complex([transcript]+[self.ribosome for i in range(n+1)], attributes = ["closed"]))

        cp_misc = [self.ribosome, transcript, protein]

        return cp_open + cp_closed + cp_misc

    def update_reactions(self, transcript, protein, component, part_id, **keywords):
        """It sets up the following reactions.

        mRNA:RBZ_n + RBZ <--> mRNA:RBZ_n_c --> mRNA:RBZ_n+1
        kf1 = kbr, kr1 = kur, kf2 = k_iso_r
        mRNA:RBZ_n --> mRNA:RBZ_0 + n RBZ + n Protein
        kf = ktl_solo
        mRNA:RBZ_n_c --> mRNA:RBZ_0_c + n RBZ + n Protein
        kf = ktl_solo
        """

        # parameter loading
        kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
        ku = component.get_parameter("ku", part_id = part_id, mechanism = self)
        k_iso = component.get_parameter("k_iso", part_id = part_id, mechanism = self)
        ktl = component.get_parameter("ktl", part_id = part_id, mechanism = self)
        max_occ = int(component.get_parameter("max_occ", part_id = part_id, mechanism = self, return_numerical = True))

        # complex species instantiation
        cp_open = []
        cp_closed = []
        for n in range(0,max_occ):
            cp_open.append(Complex([transcript]+[self.ribosome for i in range(n+1)], attributes = ["open"]))
            cp_closed.append(Complex([transcript]+[self.ribosome for i in range(n+1)], attributes = ["closed"]))

        # Reactions
        # ribosome + complex(n) <--> complex(n+1)_closed
        rxn_open_p = [Reaction.from_massaction(inputs=[self.ribosome, cp_open[n]], outputs=[cp_closed[n+1]], k_forward=kb, k_reverse = ku) for n in range(0, max_occ-1)]

        # isomerization
        # complex(n)_closed --> complex(n)
        rxn_iso = [Reaction.from_massaction(inputs=[cp_closed[n]], outputs=[cp_open[n]], k_forward=k_iso) for n in range(0, max_occ)]

        # release/translation from open and closed states
        rxn_release_open =  []
        rxn_release_closed = []
        for n in range(0,max_occ):
            rxn_temp1 = Reaction.from_massaction(inputs= [cp_open[n]], outputs=[self.ribosome for i in range(n+1)] +
                                                                               [protein for i in range(n+1)] + [transcript], k_forward=ktl)
            rxn_release_open.append(rxn_temp1)

        for n in range(1,max_occ):
            rxn_temp2 = Reaction.from_massaction(inputs= [cp_closed[n]], outputs=[self.ribosome for i in range(n)] +
                                                                                 [protein for i in range(n)] + [cp_closed[0]], k_forward=ktl)
            rxn_release_closed.append(rxn_temp2)

        # missing reactions (0 --> 0_closed and v.v. 0_closed --> 0)
        rxn_m1 = Reaction.from_massaction(inputs=[transcript, self.ribosome], outputs=[cp_closed[0]], k_forward=kb, k_reverse = ku)
        #rxn_m2 = Reaction.from_massaction(inputs=[cp_closed[0]], outputs=[transcript, self.ribosome], k_forward=kur)

        rxn_all = [rxn_m1] + rxn_iso + rxn_open_p + rxn_release_open + rxn_release_closed

        return rxn_all
